from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import attr
from aiohttp import FormData

from dis_snek.const import MISSING
from dis_snek.mixins.send import SendMixin
from dis_snek.models.application_commands import CallbackTypes
from dis_snek.models.discord_objects.message import process_message_payload
from dis_snek.models.enums import MessageFlags
from dis_snek.utils.attr_utils import define
from dis_snek.models.snowflake import to_snowflake
from dis_snek.utils.input_utils import get_args

if TYPE_CHECKING:
    from dis_snek.client import Snake
    from dis_snek.models.discord_objects.channel import TYPE_MESSAGEABLE_CHANNEL
    from dis_snek.models.discord_objects.components import ActionRow, BaseComponent
    from dis_snek.models.discord_objects.embed import Embed
    from dis_snek.models.discord_objects.guild import Guild
    from dis_snek.models.discord_objects.message import AllowedMentions, Message
    from dis_snek.models.discord_objects.user import User, Member
    from dis_snek.models.snowflake import Snowflake_Type
    from dis_snek.models.discord_objects.message import MessageReference
    from dis_snek.models.discord_objects.sticker import Sticker
    from dis_snek.models.discord_objects.role import Role


@attr.s
class Resolved:
    """Represents resolved data in an interaction

    Attributes:
        channels: A dictionary of channels mentioned in the interaction
        members: A dictionary of members mentioned in the interaction
        users: A dictionary of users mentioned in the interaction
        roles: A dictionary of roles mentioned in the interaction
    """

    channels: Dict["Snowflake_Type", "TYPE_MESSAGEABLE_CHANNEL"] = attr.ib(factory=dict)
    members: Dict["Snowflake_Type", "Member"] = attr.ib(factory=dict)
    users: Dict["Snowflake_Type", "User"] = attr.ib(factory=dict)
    roles: Dict["Snowflake_Type", "Role"] = attr.ib(factory=dict)

    @classmethod
    def from_dict(cls, client: "Snake", data: dict, guild_id: Optional["Snowflake_Type"] = None):
        new_cls = cls()

        if channels := data.get("channels"):
            for key, _channel in channels.items():
                new_cls.channels[key] = client.cache.place_channel_data(_channel)

        if members := data.get("members"):
            for key, _member in members.items():
                new_cls.members[key] = client.cache.place_member_data(
                    guild_id, {**_member, "user": {**data["users"][key]}}
                )

        if users := data.get("users"):
            for key, _user in users.items():
                new_cls.users[key] = client.cache.place_user_data(_user)

        if roles := data.get("roles"):
            for key, _role in roles.items():
                new_cls.roles[key] = client.cache.role_cache.get(to_snowflake(key))

        return new_cls


@define
class Context:
    """
    Represents the context of a command

    Attributes:
        message Message: The message associated with this context
        invoked_name str: The name of the command to be invoked
        author User: The author of the message
        channel Channel: The channel this was sent within
        guild Guild: The guild this was sent within, if not a DM
        args list: The list of arguments to be passed to the command
        kwargs dict: The list of keyword arguments to be passed

    """

    _client: "Snake" = attr.ib(default=None)
    message: "Message" = attr.ib(default=None)
    invoked_name: str = attr.ib(default=None)

    args: List = attr.ib(factory=list)

    kwargs: Dict = attr.ib(factory=dict)

    author: Union["Member", "User"] = attr.ib(default=None)
    channel: "TYPE_MESSAGEABLE_CHANNEL" = attr.ib(default=None)
    guild_id: "Snowflake_Type" = attr.ib(default=None, converter=to_snowflake)

    @property
    def guild(self):
        return self._client.cache.guild_cache.get(self.guild_id)

    @property
    def bot(self):
        """A reference to the bot instance"""
        return self._client


@define
class InteractionContext(Context, SendMixin):
    """
    Represents the context of an interaction

    !!! info "Ephemeral messages:"
        Ephemeral messages allow you to send messages that only the author of the interaction can see.
        They are best considered as `fire-and-forget`, in the sense that you cannot edit them once they have been sent.

        Should you attach a component (ie. button) to the ephemeral message,
        you will be able to edit it when responding to a button interaction.

    Attributes:
        interaction_id str: The id of the interaction
        target_id Snowflake_Type: The ID of the target, used for context menus to show what was clicked on
        deferred bool: Is this interaction deferred?
        responded bool: Have we responded to the interaction?
        ephemeral bool: Are responses to this interaction *hidden*
        resolved dict: A dictionary of discord objects mentioned within this interaction
        data dict: The raw data of this interaction
    """

    _token: str = attr.ib(default=None)
    interaction_id: str = attr.ib(default=None)
    target_id: "Snowflake_Type" = attr.ib(default=None)

    deferred: bool = attr.ib(default=False)
    responded: bool = attr.ib(default=False)
    ephemeral: bool = attr.ib(default=False)

    resolved: Resolved = attr.ib(default=Resolved())

    data: Dict = attr.ib(factory=dict)

    @classmethod
    def from_dict(cls, data: Dict, client: "Snake"):
        """Create a context object from a dictionary"""
        new_cls = cls(
            client=client,
            token=data["token"],
            interaction_id=data["id"],
            data=data,
            invoked_name=data["data"].get("name"),
            guild_id=data.get("guild_id"),
        )
        if res_data := data["data"].get("resolved"):
            new_cls.resolved = Resolved.from_dict(client, res_data, new_cls.guild_id)

        if new_cls.guild_id:
            new_cls.author = client.cache.place_member_data(new_cls.guild_id, data["member"].copy())
            client.cache.place_user_data(data["member"]["user"])
            new_cls.channel = client.cache.channel_cache.get(to_snowflake(data["channel_id"]))
        else:
            new_cls.author = client.cache.place_user_data(data["user"])
            new_cls.channel = client.cache.channel_cache.get(new_cls.author.id)

        new_cls.target_id = data["data"].get("target_id")

        new_cls.data = data
        return new_cls

    @property
    def guild(self):
        return self._client.cache.guild_cache.get(self.guild_id)

    async def defer(self, ephemeral=False) -> None:
        """
        Defers the response, showing a loading state.

        parameters:
            ephemeral: Should the response be ephemeral
        """
        if self.deferred or self.responded:
            raise Exception("You have already responded to this interaction!")

        payload = {"type": CallbackTypes.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE}
        if ephemeral:
            payload["data"] = {"flags": MessageFlags.EPHEMERAL}

        await self._client.http.post_initial_response(payload, self.interaction_id, self._token)
        self.ephemeral = ephemeral
        self.deferred = True

    async def _send_http_request(self, message_payload: Union[dict, "FormData"]) -> dict:
        if self.responded:
            message_data = await self._client.http.post_followup(message_payload, self._client.user.id, self._token)
        else:
            if isinstance(message_payload, FormData) and not self.deferred:
                await self.defer()
            if self.deferred:
                message_data = await self._client.http.edit_interaction_message(
                    message_payload, self._client.user.id, self._token
                )
                self.deferred = False
            else:
                payload = {"type": CallbackTypes.CHANNEL_MESSAGE_WITH_SOURCE, "data": message_payload}
                await self._client.http.post_initial_response(payload, self.interaction_id, self._token)
                message_data = await self._client.http.get_interaction_message(self._client.user.id, self._token)
            self.responded = True

        return message_data

    async def send(
        self,
        content: Optional[str] = None,
        embeds: Optional[Union[List[Union["Embed", dict]], Union["Embed", dict]]] = None,
        components: Optional[
            Union[List[List[Union["BaseComponent", dict]]], List[Union["BaseComponent", dict]], "BaseComponent", dict]
        ] = None,
        stickers: Optional[Union[List[Union["Sticker", "Snowflake_Type"]], "Sticker", "Snowflake_Type"]] = None,
        allowed_mentions: Optional[Union["AllowedMentions", dict]] = None,
        reply_to: Optional[Union["MessageReference", "Message", dict, "Snowflake_Type"]] = None,
        file: Optional[Union[str, "Path"]] = None,
        tts: bool = False,
        flags: Optional[Union[int, "MessageFlags"]] = None,
        ephemeral: bool = False,
    ) -> "Message":
        """
        Send a message.

        parameters:
            content: Message text content.
            embeds: Embedded rich content (up to 6000 characters).
            components: The components to include with the message.
            stickers: IDs of up to 3 stickers in the server to send in the message.
            allowed_mentions: Allowed mentions for the message.
            reply_to: Message to reference, must be from the same channel.
            filepath: Location of file to send, defaults to None.
            tts: Should this message use Text To Speech.
            flags: Message flags to apply.
            ephemeral bool: Should this message be sent as ephemeral (hidden)

        returns:
            New message object that was sent.
        """
        if ephemeral:
            flags = MessageFlags.EPHEMERAL

        return await super().send(content, embeds, components, stickers, allowed_mentions, reply_to, file, tts, flags)


@define
class ComponentContext(InteractionContext):
    custom_id: str = attr.ib(default="")
    component_type: int = attr.ib(default=0)

    values: List = attr.ib(factory=list)

    defer_edit_origin: bool = attr.ib(default=False)

    @classmethod
    def from_dict(cls, data: Dict, client: "Snake") -> "ComponentContext":
        """Create a context object from a dictionary"""
        new_cls = super().from_dict(data, client)
        new_cls.token = data["token"]
        new_cls.interaction_id = data["id"]
        new_cls.custom_id = data["data"]["custom_id"]
        new_cls.component_type = data["data"]["component_type"]
        new_cls.message = client.cache.place_message_data(data["message"])

        return new_cls

    async def defer(self, ephemeral=False, edit_origin: bool = False) -> None:
        """
        Defers the response, showing a loading state.

        parameters:
            ephemeral: Should the response be ephemeral
            edit_origin: Whether we intend to edit the original message
        """
        if self.deferred or self.responded:
            raise Exception("You have already responded to this interaction!")

        payload = {
            "type": CallbackTypes.DEFERRED_UPDATE_MESSAGE
            if edit_origin
            else CallbackTypes.DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE
        }

        if ephemeral:
            if edit_origin:
                raise ValueError("`edit_origin` and `ephemeral` are mutually exclusive")
            payload["data"] = {"flags": MessageFlags.EPHEMERAL}

        await self._client.http.post_initial_response(payload, self.interaction_id, self._token)
        self.deferred = True
        self.ephemeral = ephemeral
        self.defer_edit_origin = edit_origin

    async def edit_origin(
        self,
        content: str = None,
        embeds: List["Embed"] = None,
        components: List[Union[Dict, "ActionRow"]] = None,
        allowed_mentions: Optional[Union["AllowedMentions", dict]] = None,
        filepath: Optional[Union[str, Path]] = None,
        tts: bool = False,
    ) -> "Message":
        """
        Edits the original message of the component.

        parameters:
            content: Message text content.
            embeds: Embedded rich content (up to 6000 characters).
            components: The components to include with the message.
            allowed_mentions: Allowed mentions for the message.
            reply_to: Message to reference, must be from the same channel.
            filepath: Location of file to send, defaults to None.
            tts: Should this message use Text To Speech.

        returns:
            The message after it was edited.
        """
        if not self.responded and not self.deferred and filepath:
            # Discord doesn't allow files at initial response, so we defer then edit.
            await self.defer(edit_origin=True)

        message_payload = process_message_payload(
            content=content,
            embeds=embeds,
            components=components,
            allowed_mentions=allowed_mentions,
            file=filepath,
            tts=tts,
        )

        message_data = None
        if self.deferred:
            message_data = await self._client.http.edit_interaction_message(
                message_payload, self._client.user.id, self._token
            )
            self.deferred = False
            self.defer_edit_origin = False
        else:
            payload = {"type": CallbackTypes.UPDATE_MESSAGE, "data": message_payload}
            await self._client.http.post_initial_response(payload, self.interaction_id, self._token)
            message_data = await self._client.http.get_interaction_message(self._client.user.id, self._token)

        if message_data:
            self.message = self._client.cache.place_message_data(message_data)
            return self.message


@define
class MessageContext(Context, SendMixin):
    @classmethod
    def from_message(cls, client: "Snake", message: "Message"):
        new_cls = cls(
            client=client,
            message=message,
            author=message.author,
            channel=message.channel,
            guild_id=message.guild.id,
        )
        new_cls.args = get_args(message.content)[1:]
        return new_cls

    async def _send_http_request(self, message_payload: Union[dict, "FormData"]) -> dict:
        return await self._client.http.create_message(message_payload, self.channel.id)



from typing import TYPE_CHECKING, Any, Dict, List

from dis_snek.const import MISSING
from dis_snek.models.route import Route
from dis_snek.utils.serializer import dict_filter_missing

if TYPE_CHECKING:
    from dis_snek.models.snowflake import Snowflake_Type


class MemberRequests:
    request: Any

    async def get_member(self, guild_id: "Snowflake_Type", user_id: "Snowflake_Type") -> Dict:
        """
        Get a member of a guild by ID.

        parameters:
            guild_id: The id of the guild
            user_id: The user id to grab

        """
        return await self.request(Route("GET", f"/guilds/{guild_id}/members/{user_id}"))

    async def list_members(
        self, guild_id: "Snowflake_Type", limit: int = 1, after: "Snowflake_Type" = MISSING
    ) -> List[Dict]:
        """
        List the members of a guild.

        parameters:
            guild_id: The ID of the guild
            limit: How many members to get (max 1000)
            after: Get IDs after this snowflake

        """
        payload = dict(limit=limit)
        if after is not MISSING:
            payload["after"] = after

        return await self.request(Route("GET", f"/guilds/{guild_id}/members"), params=payload)

    async def search_guild_members(self, guild_id: "Snowflake_Type", query: str, limit: int = 1) -> List[Dict]:
        """
        Search a guild for members who's username or nickname starts with provided string.

        parameters:
            guild_id: The ID of the guild to search
            query: The string to search for
            limit: The number of members to return

        """

        return await self.request(
            Route("GET", f"/guilds/{guild_id}/members/search"), params=dict(query=query, limit=limit)
        )

    async def modify_guild_member(
        self,
        guild_id: "Snowflake_Type",
        user_id: "Snowflake_Type",
        nickname: str = MISSING,
        roles: List["Snowflake_Type"] = MISSING,
        mute: bool = MISSING,
        deaf: bool = MISSING,
        channel_id: "Snowflake_Type" = MISSING,
        reason: str = MISSING,
    ) -> Dict:
        """
        Modify attributes of a guild member.

        parameters:
            guild_id: The ID of the guild
            user_id: The ID of the user we're modifying
            nickname: Value to set users nickname to
            roles: Array of role ids the member is assigned
            mute: Whether the user is muted in voice channels. Will throw a 400 if the user is not in a voice channel
            deaf: Whether the user is deafened in voice channels
            channel_id: id of channel to move user to (if they are connected to voice)
            reason: An optional reason for the audit log
        returns:
            The updated member object
        """
        return await self.request(
            Route("PATCH", f"/guilds/{guild_id}/members/{user_id}"),
            data=dict_filter_missing(dict(nick=nickname, roles=roles, mute=mute, deaf=deaf, channel_id=channel_id)),
            reason=reason,
        )

    async def add_guild_member_role(
        self, guild_id: "Snowflake_Type", user_id: "Snowflake_Type", role_id: "Snowflake_Type", reason: str = MISSING
    ) -> None:
        """
        Adds a role to a guild member.

        parameters:
            guild_id: The ID of the guild
            user_id: The ID of the user
            role_id: The ID of the role to add
            reason: The reason for this action
        """
        return await self.request(Route("PUT", f"/guilds/{guild_id}/members/{user_id}/roles/{role_id}"))

    async def remove_guild_member_role(
        self, guild_id: "Snowflake_Type", user_id: "Snowflake_Type", role_id: "Snowflake_Type", reason: str = MISSING
    ) -> None:
        """
        Remove a role from a guild member.

        parameters:
            guild_id: The ID of the guild
            user_id: The ID of the user
            role_id: The ID of the role to remove
            reason: The reason for this action
        """
        return await self.request(
            Route("DELETE", f"/guilds/{guild_id}/members/{user_id}/roles/{role_id}"), reason=reason
        )



from typing import Dict, Any, TYPE_CHECKING, Callable, Coroutine

import aiohttp

from dis_snek.const import MISSING, GLOBAL_SCOPE

if TYPE_CHECKING:
    from dis_snek.models.command import BaseCommand
    from dis_snek.models.cooldowns import CooldownSystem, MaxConcurrency
    from dis_snek.models.snowflake import Snowflake_Type


class SnakeException(Exception):
    """Base Exception of discord-snakes"""


class BotException(SnakeException):
    """An issue occurred in the client, likely user error"""


class GatewayNotFound(SnakeException):
    """An exception that is raised when the gateway for Discord could not be found"""

    def __init__(self):
        super().__init__("Unable to find discord gateway!")


class LoginError(BotException):
    """The bot failed to login, check your token"""


class HTTPException(SnakeException):
    """A HTTP request resulted in an exception

    Attributes:
        response aiohttp.ClientResponse: The response of the HTTP request
        text str: The text of the exception, could be None
        status int: The HTTP status code
        code int: The discord error code, if one is provided
        route Route: The HTTP route that was used
    """

    def __init__(self, response: aiohttp.ClientResponse, text=MISSING, discord_code=MISSING, **kwargs):
        self.response: aiohttp.ClientResponse = response
        self.status: int = response.status
        self.code: int = discord_code
        self.text: str = text
        self.errors: Any = MISSING
        self.route = kwargs.get("route", MISSING)

        if data := kwargs.get("response_data"):
            if isinstance(data, dict):
                self.text = data.get("message", MISSING)
                self.code = data.get("code", MISSING)
                self.errors = data.get("errors", MISSING)
            else:
                self.text = data

        super().__init__(f"{self.status}|{self.response.reason}: {f'({self.code}) ' if self.code else ''}{self.text}")


class DiscordError(HTTPException):
    """A discord-side error"""


class BadRequest(HTTPException):
    """A bad request was made"""


class Forbidden(HTTPException):
    """You do not have access to this"""


class NotFound(HTTPException):
    """This resource could not be found"""


class RateLimited(HTTPException):
    """Discord is rate limiting this application"""


class TooManyChanges(SnakeException):
    """You have changed something too frequently"""


class WebSocketClosed(SnakeException):
    """The websocket was closed"""

    code: int = 0
    codes: Dict[int, str] = {
        4000: "Unknown Error",
        4001: "Unknown OpCode",
        4002: "Decode Error",
        4003: "Not Authenticated",
        4004: "Authentication Failed",
        4005: "Already Authenticated",
        4007: "Invalid seq",
        4008: "Rate limited",
        4009: "Session Timed Out",
        4010: "Invalid Shard",
        4011: "Sharding Required",
        4012: "Invalid API Version",
        4013: "Invalid Intents",
        4014: "Disallowed Intents",
    }

    def __init__(self, code: int):
        self.code = code
        super().__init__(f"The Websocket closed with code: {code} - {self.codes.get(code, 'Unknown Error')}")


class WebSocketRestart(SnakeException):
    """The websocket closed, and is safe to restart"""

    resume: bool = False

    def __int__(self, resume: bool = False):
        self.resume = resume
        super().__init__("Websocket connection closed... reconnecting")


class ExtensionException(BotException):
    """An error occurred with an extension"""


class ExtensionNotFound(ExtensionException):
    """The desired extension was not found"""


class ExtensionLoadException(ExtensionException):
    """An error occurred loading an extension"""


class ScaleLoadException(ExtensionLoadException):
    """A scale failed to load"""


class CommandException(BotException):
    """An error occurred trying to execute a command"""


class CommandOnCooldown(CommandException):
    """A command is on cooldown, and was attempted to be executed

    Attributes:
        command BaseCommand: The command that is on cooldown
        cooldown CooldownSystem: The cooldown system
    """

    def __init__(self, command: "BaseCommand", cooldown: "CooldownSystem"):
        self.command: "BaseCommand" = command
        self.cooldown: "CooldownSystem" = cooldown

        super().__init__(f"Command on cooldown... {cooldown.get_cooldown_time():.2f} seconds until reset")


class MaxConcurrencyReached(CommandException):
    """A command has exhausted the max concurrent requests."""

    def __init__(self, command: "BaseCommand", max_conc: "MaxConcurrency"):
        self.command: "BaseCommand" = command
        self.max_conc: "MaxConcurrency" = max_conc

        super().__init__(f"Command has exhausted the max concurrent requests. ({max_conc.concurrent} simultaneously)")


class CommandCheckFailure(CommandException):
    """A command check failed

    Attributes:
        command BaseCommand: The command that's check failed
        check Callable[..., Coroutine]: The check that failed
    """

    def __init__(self, command: "BaseCommand", check: Callable[..., Coroutine]):
        self.command: "BaseCommand" = command
        self.check: Callable[..., Coroutine] = check


class MessageException(BotException):
    """A message operation encountered an exception."""


class EphemeralEditException(MessageException):
    """Your bot attempted to edit an ephemeral message. This is not possible.

    Its worth noting you can edit an ephemeral message with component's `edit_origin` method"""

    def __init__(self):
        super().__init__("Ephemeral messages cannot be edited.")


class ThreadException(BotException):
    """A thread operation encountered an exception."""


class ThreadOutsideOfGuild(ThreadException):
    """A thread was attempted to be created outside of a guild."""

    def __init__(self):
        super().__init__("Threads cannot be created outside of guilds")


class InteractionException(BotException):
    """An error occurred with an interaction"""


class InteractionMissingAccess(InteractionException):
    """The bot does not have access to the specified scope"""

    def __init__(self, scope: "Snowflake_Type"):
        self.scope: "Snowflake_Type" = scope

        if scope == GLOBAL_SCOPE:
            err_msg = "Unable to sync commands global commands"
        else:
            err_msg = (
                f"Unable to sync commands for guild `{scope}` -- Ensure the bot properly added to that guild "
                f"with `application.commands` scope. "
            )

        super().__init__(err_msg)



"""
The MIT License (MIT).

Copyright (c) 2021 - present LordOfPolls

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
import logging
import re
from base64 import b64encode
from typing import Any, Dict, Union

import aiohttp  # type: ignore

from dis_snek.const import logger_name

log = logging.getLogger(logger_name)

try:
    import orjson as json
except ImportError:
    log.warning("orjson not installed, built-in json library will be used")
    import json


_quotes = {
    '"': '"',
    "‘": "’",
    "‚": "‛",
    "“": "”",
    "„": "‟",
    "⹂": "⹂",
    "「": "」",
    "『": "』",
    "〝": "〞",
    "﹁": "﹂",
    "﹃": "﹄",
    "＂": "＂",
    "｢": "｣",
    "«": "»",
    "‹": "›",
    "《": "》",
    "〈": "〉",
}
_pending_regex = r"(1.*2|[^\s]+)"
_pending_regex = _pending_regex.replace("1", f"[{''.join([k for k in _quotes.keys()])}]")
_pending_regex = _pending_regex.replace("2", f"[{''.join([k for k in _quotes.values()])}]")

arg_parse = re.compile(_pending_regex)
white_space = re.compile(r"\s+")
initial_word = re.compile(r"^([^\s]+)\s*?")


class OverriddenJson:
    @staticmethod
    def dumps(*args, **kwargs) -> str:
        data = json.dumps(*args, **kwargs)
        if isinstance(data, bytes):
            data = data.decode("utf-8")
        return data

    @staticmethod
    def loads(*args, **kwargs) -> dict:
        return json.loads(*args, **kwargs)


async def response_decode(response: aiohttp.ClientResponse) -> Union[Dict[str, Any], str]:
    """
    Return the response text in its correct format, be it dict, or string.

    :param response: the aiohttp response
    :return: the response text field in its correct type
    """
    text = await response.text(encoding="utf-8")

    if response.headers.get("content-type") == "application/json":
        return OverriddenJson.loads(text)
    return text


def get_args(text: str):
    """
    Get arguments from an input text.

    :param text: The text to process
    :return: A list of words
    """
    return arg_parse.findall(text)


def get_first_word(text: str):
    """
    Get a the first word in a string, regardless of whitespace type.
    :param text: The text to process
    :return: The requested word
    """
    found = initial_word.findall(text)
    if len(found) == 0:
        return None
    return found[0]


def _get_mime_type_for_image(data: bytes):
    # taken from d.py, alternative is to use libmagic, which would require users to install libs
    if data.startswith(b"\x89\x50\x4E\x47\x0D\x0A\x1A\x0A"):
        return "image/png"
    elif data[0:3] == b"\xff\xd8\xff" or data[6:10] in (b"JFIF", b"Exif"):
        return "image/jpeg"
    elif data.startswith((b"\x47\x49\x46\x38\x37\x61", b"\x47\x49\x46\x38\x39\x61")):
        return "image/gif"
    elif data.startswith(b"RIFF") and data[8:12] == b"WEBP":
        return "image/webp"
    else:
        raise ValueError("Unsupported image type given")


def _bytes_to_base64_data(data: bytes) -> str:
    mime = _get_mime_type_for_image(data)
    return f"data:{mime};base64,{b64encode(data).decode('ascii')}"



"""
These are events dispatched by Discord. This is intended as a reference so you know what data to expect for each event

!!! Hint "Example Usage:"
    The event classes outlined here are in `CamelCase` to comply with Class naming convention, however the event names
    are actually in `lower_case_with_underscores` so your listeners should be named as following:

    ```python
    @bot.event
    def on_ready():
        # ready events pass no data, so dont have params
        print("Im ready!")

    @bot.event
    def on_guild_create(event):
        # guild_create events pass a guild object, expect a single param
        print(f"{event.guild.name} created")
    ```
"""

from typing import TYPE_CHECKING, Any, List, Union

import attr

from dis_snek.models.events.internal import BaseEvent, GuildEvent

if TYPE_CHECKING:
    from dis_snek.models.discord_objects.guild import Guild
    from dis_snek.models.discord_objects.channel import BaseChannel
    from dis_snek.models.discord_objects.message import Message
    from dis_snek.models.timestamp import Timestamp
    from dis_snek.models.discord_objects.user import Member, User, BaseUser
    from dis_snek.models.snowflake import Snowflake_Type
    from dis_snek.models.discord_objects.emoji import Emoji
    from dis_snek.models.discord_objects.role import Role
    from dis_snek.models.discord_objects.sticker import Sticker


@attr.s(slots=True)
class RawGatewayEvent(BaseEvent):
    """An event dispatched from the gateway. Holds the raw dict that the gateway dispatches"""

    data: dict = attr.ib(factory=dict)
    """Raw Data from the gateway"""


@attr.s(slots=True)
class ChannelCreate(BaseEvent):
    """Dispatched when a channel is created."""

    channel: "BaseChannel" = attr.ib(metadata={"docs": "The channel this event is dispatched from"})


@attr.s(slots=True)
class ChannelUpdate(ChannelCreate):
    """Dispatched when a channel is updated"""


@attr.s(slots=True)
class ChannelDelete(ChannelCreate):
    """Dispatched when a channel is deleted"""


@attr.s(slots=True)
class ChannelPinsUpdate(ChannelCreate):
    """Dispatched when a channel's pins are updated"""

    last_pin_timestamp: "Timestamp" = attr.ib()
    """The time at which the most recent pinned message was pinned"""


@attr.s(slots=True)
class ThreadCreate(BaseEvent):
    """Dispatched when a thread is created."""

    thread: Any = attr.ib(
        metadata={"docs": "The thread this event is dispatched from"}
    )  # TODO: Replace this with a thread object type


@attr.s(slots=True)
class ThreadUpdate(ThreadCreate):
    """Dispatched when a thread is updated"""


@attr.s(slots=True)
class ThreadDelete(ThreadCreate):
    """Dispatched when a thread is deleted"""


@attr.s(slots=True)
class ThreadListSync(BaseEvent):
    """Dispatched when gaining access to a channel, contains all active threads in that channel"""

    channel_ids: List["Snowflake_Type"] = attr.ib()
    """The parent channel ids whose threads are being synced. If omitted, then threads were synced for the entire guild. This array may contain channel_ids that have no active threads as well, so you know to clear that data."""
    threads: List["BaseChannel"] = attr.ib()
    """all active threads in the given channels that the current user can access"""
    members: List["Member"] = attr.ib()
    """all thread member objects from the synced threads for the current user, indicating which threads the current user has been added to"""


@attr.s(slots=True)
class ThreadMemberUpdate(ThreadCreate):
    """Dispatched when the thread member object for the current user is updated.

    ??? info "Note from Discord"
        This event is documented for completeness, but unlikely to be used by most bots. For bots, this event largely is just a signal that you are a member of the thread
    """

    member: "Member" = attr.ib()
    """The member who was added"""


@attr.s(slots=True)
class ThreadMembersUpdate(BaseEvent):
    """Dispatched when anyone is added or removed from a thread."""

    id: "Snowflake_Type" = attr.ib()
    """The ID of the thread"""
    member_count: int = attr.ib(default=50)
    """the approximate number of members in the thread, capped at 50"""
    added_members: List["Member"] = attr.ib(factory=list)
    """Users added to the thread"""
    removed_member_ids: List["Snowflake_Type"] = attr.ib(factory=list)
    """Users removed from the thread"""


@attr.s(slots=True)
class GuildCreate(BaseEvent):
    """Dispatched when a guild is created."""

    guild: "Guild" = attr.ib()
    """The guild that was created"""


@attr.s(slots=True)
class GuildUpdate(BaseEvent):
    """Dispatched when a guild is updated."""

    before: "Guild" = attr.ib()
    """Guild before this event"""
    after: "Guild" = attr.ib()
    """Guild after this event"""


@attr.s(slots=True)
class GuildDelete(BaseEvent, GuildEvent):
    """Dispatched when a guild becomes unavailable or user left/removed."""

    unavailable: bool = attr.ib(default=False)
    """If this event was triggered due to an outage"""


@attr.s(slots=True)
class GuildBanAdd(BaseEvent, GuildEvent):
    """Dispatched when someone was banned from a guild"""

    user: "BaseUser" = attr.ib(metadata={"docs": "The user"})


@attr.s(slots=True)
class GuildBanRemove(GuildBanAdd):
    """Dispatched when a users ban is removed"""


@attr.s(slots=True)
class GuildEmojisUpdate(BaseEvent, GuildEvent):
    """Dispatched when a guild's emojis are updated."""

    before: List["Emoji"] = attr.ib(factory=list)
    """List of emoji before this event"""
    after: List["Emoji"] = attr.ib(factory=list)
    """List of emoji after this event"""


@attr.s(slots=True)
class GuildStickersUpdate(BaseEvent, GuildEvent):
    """Dispatched when a guild's stickers are updated."""

    before: List["Sticker"] = attr.ib(factory=list)
    """List of stickers from before this event"""
    after: List["Sticker"] = attr.ib(factory=list)
    """List of stickers from after this event"""


@attr.s(slots=True)
class MemberAdd(BaseEvent, GuildEvent):
    """Dispatched when a member is added to a guild."""

    member: "Member" = attr.ib(metadata={"docs": "The member who was added"})


@attr.s(slots=True)
class MemberRemove(MemberAdd):
    """Dispatched when a member is removed from a guild."""


@attr.s(slots=True)
class MemberUpdate(BaseEvent, GuildEvent):
    """Dispatched when a member is updated."""

    # todo: It would be better to just have a member, before, and after field. but for now im mirroring the api

    before: "Member" = attr.ib()
    """The state of the member before this event"""
    after: "Member" = attr.ib()
    """The state of the member after this event"""


@attr.s(slots=True)
class GuildRoleCreate(BaseEvent, GuildEvent):
    """Dispatched when a role is created."""

    role: "Role" = attr.ib()
    """The created role"""


@attr.s(slots=True)
class GuildRoleUpdate(BaseEvent, GuildEvent):
    """Dispatched when a role is updated."""

    before: "Role" = attr.ib()
    """The role before this event"""
    after: "Role" = attr.ib()
    """The role after this event"""


@attr.s(slots=True)
class GuildRoleDelete(BaseEvent, GuildEvent):
    """Dispatched when a guild role is deleted"""

    role_id: "Snowflake_Type" = attr.ib()
    """The ID of the deleted role"""


@attr.s(slots=True)
class GuildMembersChunk(BaseEvent, GuildEvent):
    """Sent in response to Guild Request Members. You can use the `chunk_index` and `chunk_count` to calculate how many chunks are left for your request."""

    chunk_index: int = attr.ib()
    """The chunk index in the expected chunks for this response (0 <= chunk_index < chunk_count)"""
    chunk_count: int = attr.ib()
    """the total number of expected chunks for this response"""
    presences: List = attr.ib()
    """if passing true to `REQUEST_GUILD_MEMBERS`, presences of the returned members will be here"""
    nonce: str = attr.ib()
    """The nonce used in the request, if any"""
    members: List["Member"] = attr.ib(factory=list)
    """A list of members"""


@attr.s(slots=True)
class IntegrationCreate(BaseEvent):
    """Dispatched when a guild integration is created"""

    integration: Any = attr.ib()  # TODO: Replace this with a integration object type


@attr.s(slots=True)
class IntegrationUpdate(BaseEvent, GuildEvent):
    """Dispatched when a guild integration is updated"""

    before: Any = attr.ib()  # TODO: Replace this with a integration object type
    """The integration before this event"""
    after: Any = attr.ib()  # TODO: Replace this with a integration object type
    """The integration after this event"""


@attr.s(slots=True)
class IntegrationDelete(BaseEvent, GuildEvent):
    """Dispatched when a guild integration is deleted"""

    id: "Snowflake_Type" = attr.ib()
    """The ID of the integration"""
    application_id: "Snowflake_Type" = attr.ib(default=None)
    """The ID of the bot/application for this integration"""


@attr.s(slots=True)
class GuildIntegrationsUpdate(BaseEvent, GuildEvent):
    """Dispatched when a guild integration is updated"""


@attr.s(slots=True)
class InviteCreate(BaseEvent):
    """Dispatched when a guild invite is created"""

    invite: Any = attr.ib()  # TODO: Replace this with a invite object type


@attr.s(slots=True)
class MessageCreate(BaseEvent):
    """Dispatched when a message is created"""

    message: "Message" = attr.ib()


@attr.s(slots=True)
class MessageUpdate(BaseEvent):
    """Dispatched when a message is edited"""

    before: "Message" = attr.ib()
    """The message before this event was created"""
    after: "Message" = attr.ib()
    """The message after this event was created"""


@attr.s(slots=True)
class MessageDelete(BaseEvent):
    """Dispatched when a message is deleted."""

    message: "Message" = attr.ib()


@attr.s(slots=True)
class MessageDeleteBulk(BaseEvent, GuildEvent):
    """Dispatched when multiple messages are deleted at once."""

    channel_id: "Snowflake_Type" = attr.ib()
    """The ID of the channel these were deleted in"""
    ids: List["Snowflake_Type"] = attr.ib(factory=list)
    """A list of message snowflakes"""


@attr.s(slots=True)
class MessageReactionAdd(BaseEvent):
    """Dispatched when a reaction is added to a message."""

    message: "Message" = attr.ib(metadata={"docs": "The message that was reacted to"})
    emoji: "emoji" = attr.ib(metadata={"docs": "The emoji that was added to the message"})
    author: Union["Member", "User"] = attr.ib(metadata={"docs": "The user who added the reaction"})


@attr.s(slots=True)
class MessageReactionRemove(MessageReactionAdd):
    """Dispatched when a reaction is removed"""


@attr.s(slots=True)
class MessageReactionRemoveAll(BaseEvent, GuildEvent):
    """Dispatched when all reactions are removed from a message"""

    message: "Message" = attr.ib()
    """The message that was reacted to"""


@attr.s(slots=True)
class PresenceUpdate(BaseEvent):
    """A user's presence has changed"""

    user: "User" = attr.ib()
    """The user in question"""
    status: str = attr.ib()
    """'Either `idle`, `dnd`, `online`, or `offline`'"""
    activities: List = attr.ib()
    """The users current activities"""
    client_status: dict = attr.ib()
    """What platform the user is reported as being on"""


@attr.s(slots=True)
class StageInstanceCreate(BaseEvent):
    """Dispatched when a stage instance is created"""

    stage_instance: Any = attr.ib(
        metadata={"docs": "The stage instance"}
    )  # TODO: Replace this with a stage instance object type.


@attr.s(slots=True)
class StageInstanceDelete(StageInstanceCreate):
    """Dispatched when a stage instance is deleted"""


@attr.s(slots=True)
class StageInstanceUpdate(BaseEvent):
    """Dispatched when a stage instance is updated"""

    before: Any = attr.ib()  # TODO: Replace this
    """The stage instance before this event was created"""
    after: Any = attr.ib()  # TODO: Replace this
    """The stage instance after this event was created"""


@attr.s(slots=True)
class TypingStart(BaseEvent):
    """Dispatched when a user starts typing"""

    author: Union["User", "Member"] = attr.ib()
    """The user who started typing"""
    channel: "BaseChannel" = attr.ib()
    """The channel typing is in"""
    guild: "Guild" = attr.ib()
    """The ID of the guild this typing is in"""
    timestamp: "Timestamp" = attr.ib()
    """unix time (in seconds) of when the user started typing"""


@attr.s(slots=True)
class WebhooksUpdate(BaseEvent, GuildEvent):
    """Dispatched when a guild channel webhook is created, updated, or deleted"""

    # Discord doesnt sent the webhook object for this event, for some reason
    channel_id: "Snowflake_Type" = attr.ib()
    """The ID of the webhook was updated"""


@attr.s(slots=True)
class InteractionCreate(BaseEvent):
    """Dispatched when a user uses an Application Command"""

    interaction: dict = attr.ib()



import asyncio
import copy
import importlib.util
import inspect
import logging
import sys
import traceback
from random import randint
from typing import TYPE_CHECKING, Callable, Coroutine, Dict, List, Optional, Union

import aiohttp

from dis_snek.const import logger_name, GLOBAL_SCOPE, events, MISSING
from dis_snek.errors import (
    GatewayNotFound,
    SnakeException,
    WebSocketClosed,
    WebSocketRestart,
    BotException,
    ScaleLoadException,
    ExtensionLoadException,
    ExtensionNotFound,
    Forbidden,
    InteractionMissingAccess,
)
from dis_snek.gateway import WebsocketClient
from dis_snek.http_client import HTTPClient
from dis_snek.models.application_commands import (
    InteractionCommand,
    SlashCommand,
    OptionTypes,
    SubCommand,
)
from dis_snek.models.command import MessageCommand, BaseCommand
from dis_snek.models.context import ComponentContext, InteractionContext, MessageContext
from dis_snek.models.discord_objects.activity import Activity
from dis_snek.models.discord_objects.application import Application
from dis_snek.models.discord_objects.channel import BaseChannel
from dis_snek.models.discord_objects.guild import Guild
from dis_snek.models.discord_objects.message import Message
from dis_snek.models.discord_objects.sticker import StickerPack, Sticker
from dis_snek.models.discord_objects.user import SnakeBotUser, User, Member
from dis_snek.models.enums import ComponentTypes, Intents, InteractionTypes, Status, ActivityType
from dis_snek.models.events import RawGatewayEvent, MessageCreate
from dis_snek.models.listener import Listener, listen
from dis_snek.models.scale import Scale
from dis_snek.models.snowflake import to_snowflake
from dis_snek.models.timestamp import Timestamp
from dis_snek.smart_cache import GlobalCache
from dis_snek.utils.cache import TTLCache
from dis_snek.utils.input_utils import get_first_word, get_args
from dis_snek.utils.misc_utils import wrap_partial

if TYPE_CHECKING:
    from dis_snek.models.snowflake import Snowflake_Type


log = logging.getLogger(logger_name)


class Snake:
    """
    The bot client.

    note:
        By default, all non-privileged intents will be enabled

    Attributes:
        intents Union[int, Intents]: The intents to use
        loop: An event loop to use, normally leave this blank
        default_prefix str: The default_prefix to use for message commands, defaults to `.`
        get_prefix Callable[..., Coroutine]: A coroutine that returns a string to determine prefixes
        sync_interactions bool: Should application commands be synced with discord?
        asyncio_debug bool: Enable asyncio debug features
        message_cache_ttl int: How long a message will remain in the cache, set to `None` to disable cache expiry
        message_cache_size int: The maximum number of messages that may be stored in the cache, set to `None` to not limit cache size

    !!! note
        Setting message_cache_size to None is not recommended, as it could result in extremely high memory usage, we suggest a sane limit.

    """

    def __init__(
        self,
        intents: Union[int, Intents] = Intents.DEFAULT,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        default_prefix: str = ".",
        get_prefix: Callable[..., Coroutine] = MISSING,
        sync_interactions: bool = False,
        asyncio_debug: bool = False,
        message_cache_ttl: Optional[int] = 600,
        message_cache_limit: Optional[int] = 250,
    ):

        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop() if loop is None else loop

        # Configuration

        if asyncio_debug:
            log.warning("Asyncio Debug is enabled, Your log will contain additional errors and warnings")
            import tracemalloc

            tracemalloc.start()
            self.loop.set_debug(True)

        self.intents = intents
        """The intents in use"""
        self.sync_interactions = sync_interactions
        """Should application commands be synced"""
        self.default_prefix = default_prefix
        """The default prefix to be used for message commands"""
        self.get_prefix = get_prefix if get_prefix is not MISSING else self.get_prefix
        """A coroutine that returns a prefix, for dynamic prefixes"""

        # resources

        self.http: HTTPClient = HTTPClient(loop=self.loop)
        """The HTTP client to use when interacting with discord endpoints"""
        self.ws: WebsocketClient = MISSING
        """The websocket collection for the Discord Gateway."""

        # flags
        self._ready = False
        self._closed = False
        self._guild_event = asyncio.Event()
        self.guild_event_timeout = 3
        """How long to wait for guilds to be cached"""

        # caches

        self.cache: GlobalCache = GlobalCache(self)

        if message_cache_limit is None and message_cache_ttl is None:
            log.warning("NO MESSAGE CACHE LIMITS ARE ACTIVE! This is not recommended")
            self.cache.message_cache = dict()
        else:
            self.cache.message_cache = TTLCache(hard_limit=message_cache_limit, ttl=message_cache_ttl or float("inf"))

        self._user: SnakeBotUser = MISSING
        self._app: Application = MISSING

        # collections

        self.commands: Dict[str, MessageCommand] = {}
        """A dicitonary of registered commands: `{name: command}`"""
        self.interactions: Dict["Snowflake_Type", Dict[str, InteractionCommand]] = {}
        """A dictionary of registered application commands: `{cmd_id: command}`"""
        self._interaction_scopes: Dict["Snowflake_Type", "Snowflake_Type"] = {}
        self.__extensions = {}
        self.scales = {}
        """A dictionary of mounted Scales"""
        self.listeners: Dict[str, List] = {}

    @property
    def is_closed(self) -> bool:
        """Is the bot closed?"""
        return self._closed

    @property
    def is_ready(self):
        return self._ready

    @property
    def latency(self) -> float:
        """Returns the latency of the websocket connection"""
        return self.ws.latency

    @property
    def user(self) -> SnakeBotUser:
        """Returns the bot's user"""
        return self._user

    @property
    def app(self) -> Application:
        """Returns the bots application"""
        return self._app

    @property
    def owner(self) -> Optional["User"]:
        """Returns the bot's owner'"""
        try:
            return self.app.owner
        except TypeError:
            return MISSING

    @property
    def guilds(self) -> List["Guild"]:
        return self.user.guilds

    async def get_prefix(self, message: Message) -> str:
        """A method to get the bot's default_prefix, can be overridden to add dynamic prefixes.

        !!! note
            To easily override this method, simply use the `get_prefix` parameter when instantiating the client

        Args:
            message: A message to determine the prefix from.

        Returns:
            A string to use as a prefix, by default will return `client.default_prefix`
        """
        return self.default_prefix

    async def login(self, token):
        """
        Login to discord

        Args:
            token str: Your bot's token
        """
        # i needed somewhere to put this call,
        # login will always run after initialisation
        # so im gathering commands here
        self._gather_commands()

        log.debug(f"Logging in with token: {token}")
        me = await self.http.login(token.strip())
        self._user = SnakeBotUser.from_dict(me, self)
        self.cache.place_user_data(me)
        self._app = Application.from_dict(await self.http.get_current_bot_information(), self)
        self.dispatch(events.Login())
        await self._ws_connect()

    async def _ws_connect(self):
        params = {
            "http": self.http,
            "dispatch": self.dispatch,
            "intents": self.intents,
            "resume": False,
            "session_id": None,
            "sequence": None,
        }
        while not self.is_closed:
            log.info(f"Attempting to {'re' if params['resume'] else ''}connect to gateway...")

            try:
                self.ws = await WebsocketClient.connect(**params)

                await self.ws.run()
            except WebSocketRestart as ex:
                # internally requested restart
                self.dispatch(events.Disconnect())
                if ex.resume:
                    params.update(resume=True, session_id=self.ws.session_id, sequence=self.ws.sequence)
                    continue
                params.update(resume=False, session_id=None, sequence=None)

            except (OSError, GatewayNotFound, aiohttp.ClientError, asyncio.TimeoutError, WebSocketClosed) as ex:
                self.dispatch(events.Disconnect())

                if isinstance(ex, WebSocketClosed):
                    if ex.code == 1000:
                        # clean close
                        return
                    elif ex.code == 4011:
                        raise SnakeException("Your bot is too large, you must use shards") from None
                    elif ex.code == 4013:
                        raise SnakeException("Invalid Intents have been passed") from None
                    elif ex.code == 4014:
                        raise SnakeException(
                            "You have requested privileged intents that have not been enabled or approved. Check the developer dashboard"
                        ) from None
                    raise

                if isinstance(ex, OSError) and ex.errno in (54, 10054):
                    print("should reconnect")
                    params.update(resume=True, session_id=self.ws.session_id, sequence=self.ws.sequence)
                    continue
                params.update(resume=False, session_id=None, sequence=None)

            except Exception as e:
                self.dispatch(events.Disconnect())
                log.error("".join(traceback.format_exception(type(e), e, e.__traceback__)))
                params.update(resume=False, session_id=None, sequence=None)

            await asyncio.sleep(randint(1, 5))

    def _queue_task(self, coro, event, *args, **kwargs):
        async def _async_wrap(_coro, _event, *_args, **_kwargs):
            try:
                if len(_event.__attrs_attrs__) == 1:
                    await _coro()
                else:
                    await _coro(_event, *_args, **_kwargs)
            except asyncio.CancelledError:
                pass
            except Exception:
                raise

        wrapped = _async_wrap(coro, event, *args, **kwargs)

        return asyncio.create_task(wrapped, name=f"snake:: {event.resolved_name}")

    def start(self, token):
        """
        Start the bot.

        info:
            This is the recommended method to start the bot

        Args:
            token str: Your bot's token
        """
        try:
            self.loop.run_until_complete(self.login(token))
        except KeyboardInterrupt:
            self.loop.run_until_complete(self.stop())

    async def stop(self):
        log.debug("Stopping the bot.")
        await self.ws.close()

    def dispatch(self, event: events.BaseEvent, *args, **kwargs):
        """
        Dispatch an event.

        Args:
            event: The event to be dispatched.
        """
        log.debug(f"Dispatching Event: {event.resolved_name}")
        listeners = self.listeners.get(event.resolved_name, [])
        for _listen in listeners:
            try:
                self._queue_task(_listen, event, *args, **kwargs)
            except Exception as e:
                raise BotException(f"An error occurred attempting during {event.resolved_name} event processing") from e

    def add_listener(self, listener: Listener):
        """
        Add a listener for an event, if no event is passed, one is determined

        Args:
            coro Listener: The listener to add to the client
        """
        if listener.event not in self.listeners:
            self.listeners[listener.event] = []
        self.listeners[listener.event].append(listener)

    async def _init_interactions(self) -> None:
        """
        Initialise slash commands.

        If `sync_interactions` this will submit all registered slash commands to discord.
        Otherwise, it will get the list of interactions and cache their scopes.
        """
        # allow for cogs and main to share the same decorator

        if self.sync_interactions:
            await self.synchronise_interactions()
        else:
            await self._cache_interactions(warn_missing=True)

    async def _cache_interactions(self, warn_missing: bool = False):
        """Get all interactions used by this bot and cache them."""
        bot_scopes = set(g.id for g in self.cache.guild_cache.values())
        bot_scopes.add(GLOBAL_SCOPE)

        # Match all interaction is registered with discord's data.
        for scope in self.interactions:
            bot_scopes.discard(scope)
            try:
                remote_cmds = await self.http.get_interaction_element(self.user.id, scope)
            except Forbidden as e:
                raise InteractionMissingAccess(scope) from e

            remote_cmds = {cmd_data["name"]: cmd_data for cmd_data in remote_cmds}
            for cmd in self.interactions[scope].values():
                cmd_data = remote_cmds.pop(cmd.name, MISSING)
                if cmd_data is MISSING:
                    if warn_missing:
                        log.error(
                            f'Detected yet to sync slash command "/{cmd.name}" for scope '
                            f"{'global' if scope == GLOBAL_SCOPE else scope}"
                        )
                    continue

                self._interaction_scopes[str(cmd_data["id"])] = scope
                cmd.cmd_id = str(cmd_data["id"])

            if warn_missing:
                for cmd_data in remote_cmds.values():
                    log.error(
                        f"Detected unimplemented slash command \"/{cmd_data['name']}\" for scope "
                        f"{'global' if scope == GLOBAL_SCOPE else scope}"
                    )

        # Remaining guilds that bot is in but, no interaction is registered
        for scope in bot_scopes:
            try:
                remote_cmds = await self.http.get_interaction_element(self.user.id, scope)
            except Forbidden:
                # We will just assume they don't want application commands in this guild.
                log.debug(f"Bot was not invited to guild {scope} with `application.commands` scope")
                continue

            for cmd_data in remote_cmds:
                self._interaction_scopes[str(cmd_data["id"])] = scope
                if warn_missing:
                    log.error(
                        f"Detected unimplemented slash command \"/{cmd_data['name']}\" for scope "
                        f"{'global' if scope == GLOBAL_SCOPE else scope}"
                    )

    def _gather_commands(self):
        """Gathers commands from __main__ and self"""

        def process(_cmds):

            for func in _cmds:
                if isinstance(func, InteractionCommand):
                    self.add_interaction(func)
                if isinstance(func, MessageCommand):
                    self.add_message_command(func)
                if isinstance(func, Listener):
                    self.add_listener(func)

            log.debug(f"{len(_cmds)} commands have been loaded")

        process(
            [obj for _, obj in inspect.getmembers(sys.modules["__main__"]) if isinstance(obj, (BaseCommand, Listener))]
        )
        process(
            [wrap_partial(obj, self) for _, obj in inspect.getmembers(self) if isinstance(obj, (BaseCommand, Listener))]
        )

    def add_interaction(self, command: InteractionCommand):
        """
        Add a slash command to the client.

        Args:
            command InteractionCommand: The command to add
        """
        if command.scope not in self.interactions:
            self.interactions[command.scope] = {}
        elif command.resolved_name in self.interactions[command.scope]:
            old_cmd = self.interactions[command.scope][command.name]
            raise ValueError(f"Duplicate Command! {old_cmd.scope}::{old_cmd.resolved_name}")

        if isinstance(command, SubCommand):
            if existing_sub := self.interactions[command.scope].get(command.name):
                if command.group_name:
                    existing_index = next(
                        (
                            index
                            for (index, val) in enumerate(existing_sub.options)
                            if val["name"] == command.group_name and val["type"] == 2
                        ),
                        None,
                    )
                    if existing_index is not None:
                        data = command.child_to_dict()
                        existing_sub.options[existing_index]["options"] += data["options"]
                        existing_sub.subcommand_callbacks[command.resolved_name] = command.callback
                        self.interactions[command.scope][command.name] = existing_sub
                        return

                existing_sub.options += [command.child_to_dict()]
                existing_sub.subcommand_callbacks[command.resolved_name] = command.callback
                command = existing_sub

            else:
                command.options = [command.child_to_dict()]
                command.subcommand_callbacks[command.resolved_name] = command.callback
                command.callback = command.subcommand_call

        self.interactions[command.scope][command.name] = command

    def add_message_command(self, command: MessageCommand):
        """
        Add a message command to the client.

        Args:
            command InteractionCommand: The command to add
        """
        if command.name not in self.commands:
            self.commands[command.name] = command
            return
        raise ValueError(f"Duplicate Command! Multiple commands share the name `{command.name}`")

    async def synchronise_interactions(self) -> None:
        """Synchronise registered interactions with discord

        One flaw of this is it cant determine if context menus need updating,
        as discord isn't returning that data on get req, so they are unnecessarily updated"""

        # first we need to make sure our local copy of cmd_ids is up-to-date
        await self._cache_interactions()
        cmd_scopes = [k for k in self.interactions.keys()]
        guild_perms = {}

        for cmd_scope in cmd_scopes:
            try:
                cmds_resp_data = await self.http.get_interaction_element(self.user.id, cmd_scope)
                need_to_sync = False
                cmds_to_sync = []

                for local_cmd in self.interactions[cmd_scope].values():
                    # try and find remote equiv of this command
                    remote_cmd = next((v for v in cmds_resp_data if v["id"] == local_cmd.cmd_id), None)
                    local_cmd = local_cmd.to_dict()
                    cmds_to_sync.append(local_cmd)

                    if (
                        not remote_cmd
                        or local_cmd["name"] != remote_cmd["name"]
                        or local_cmd.get("description", "") != remote_cmd.get("description", "")
                        or local_cmd.get("default_permission", True) != remote_cmd.get("default_permission", True)
                        or local_cmd.get("options") != remote_cmd.get("options")
                    ):  # if command local data doesnt match remote, a change has been made, sync it
                        need_to_sync = True

                if need_to_sync:
                    log.debug(f"Updating {len(cmds_to_sync)} commands in {cmd_scope}")
                    cmd_sync_resp = await self.http.post_interaction_element(
                        self.user.id, cmds_to_sync, guild_id=cmd_scope
                    )
                    # cache cmd_ids and their scopes
                    for cmd_data in cmd_sync_resp:
                        self.interactions[cmd_scope][cmd_data["name"]].cmd_id = str(cmd_data["id"])
                        self._interaction_scopes[cmd_data["id"]] = cmd_scope
                else:
                    log.debug(f"{cmd_scope} is already up-to-date")

                for local_cmd in self.interactions[cmd_scope].values():
                    if not local_cmd.permissions:
                        continue

                    for perm_scope, perms in local_cmd.permissions.items():
                        if perm_scope not in guild_perms:
                            guild_perms[perm_scope] = []
                        guild_perms[perm_scope].append(
                            {"id": local_cmd.cmd_id, "permissions": [perm.to_dict() for perm in perms]}
                        )

                for perm_scope in guild_perms:
                    log.debug(f"Updating {len(guild_perms[perm_scope])} command permissions in {perm_scope}")
                    await self.http.batch_edit_application_command_permissions(
                        application_id=self.user.id, scope=perm_scope, data=guild_perms[perm_scope]
                    )
            except Forbidden as e:
                raise InteractionMissingAccess(cmd_scope) from e

    async def get_context(
        self, data: Union[dict, Message], interaction: bool = False
    ) -> Union[MessageContext, InteractionContext, ComponentContext]:
        """
        Return a context object based on data passed

        note:
            If you want to use custom context objects, this is the method to override. Your replacement must take the same arguments as this, and return a Context-like object.

        Args:
            data: The data of the event
            interaction: Is this an interaction or not?

        returns:
            Context object
        """
        # this line shuts up IDE warnings
        cls: Union[MessageContext, ComponentContext, InteractionContext]

        if interaction:
            if data["type"] == InteractionTypes.MESSAGE_COMPONENT:
                cls = ComponentContext.from_dict(data, self)
            else:
                cls = InteractionContext.from_dict(data, self)

        else:
            cls = MessageContext.from_message(self, data)
        return cls

    @listen("raw_interaction_create")
    async def _dispatch_interaction(self, event: RawGatewayEvent) -> None:
        """
        Identify and dispatch interaction of slash commands or components.

        Args:
            raw interaction event
        """
        # Yes this is temporary, im just blocking out the basic logic
        interaction_data = event.data

        if interaction_data["type"] in (InteractionTypes.PING, InteractionTypes.APPLICATION_COMMAND):
            # Slash Commands
            interaction_id = interaction_data["data"]["id"]
            name = interaction_data["data"]["name"]
            invoked_name = name
            scope = self._interaction_scopes.get(str(interaction_id))

            kwargs = {}
            # todo: redo this logic properly. This is a mess and you know it, Polls
            if options := interaction_data["data"].get("options"):
                if options[0]["type"] in (OptionTypes.SUB_COMMAND, OptionTypes.SUB_COMMAND_GROUP):
                    # subcommand, process accordingly
                    if options[0]["type"] == OptionTypes.SUB_COMMAND:
                        invoked_name = f"{name} {options[0]['name']}"
                    else:
                        invoked_name = (
                            f"{name} {options[0]['name']} "
                            f"{next(x for x in options[0]['options'] if x['type'] == OptionTypes.SUB_COMMAND)['name']}"
                        )
                        options = options[0]["options"][0].get("options")

                for option in options:
                    if option["type"] not in (OptionTypes.SUB_COMMAND, OptionTypes.SUB_COMMAND_GROUP):
                        kwargs[option["name"]] = option.get("value")

            if scope in self.interactions:
                command: SlashCommand = self.interactions[scope][name]
                print(f"{command.scope} :: {command.name} should be called")

                ctx = await self.get_context(interaction_data, True)
                ctx.invoked_name = invoked_name
                ctx.kwargs = kwargs
                ctx.args = kwargs.values()

                await command(ctx, **kwargs)
            else:
                log.error(f"Unknown cmd_id received:: {interaction_id} ({name})")

        elif interaction_data["type"] == InteractionTypes.MESSAGE_COMPONENT:
            # Buttons, Selects, ContextMenu::Message
            ctx = await self.get_context(interaction_data, True)
            component_type = interaction_data["data"]["component_type"]

            self.dispatch(events.Component(ctx))
            if component_type == ComponentTypes.BUTTON:
                self.dispatch(events.Button(ctx))
            if component_type == ComponentTypes.SELECT:
                self.dispatch(events.Select(ctx))

        else:
            raise NotImplementedError(f"Unknown Interaction Received: {interaction_data['type']}")

    @listen("message_create")
    async def _dispatch_msg_commands(self, event: MessageCreate):
        """Determine if a command is being triggered, and dispatch it."""
        message = event.message

        if not message.author.bot:
            prefix = await self.get_prefix(message)

            if message.content.startswith(prefix):
                invoked_name = get_first_word(message.content.removeprefix(prefix))
                command = self.commands.get(invoked_name)
                if command and command.enabled:
                    context = await self.get_context(message)
                    context.invoked_name = invoked_name
                    await command(context)

    @listen()
    async def _on_raw_message_create(self, event: RawGatewayEvent) -> None:
        """
        Automatically convert MESSAGE_CREATE event data to the object.

        Args:
            event: raw message event
        """
        msg = self.cache.place_message_data(event.data)
        if not msg._guild_id and event.data.get("guild_id"):
            msg._guild_id = event.data["guild_id"]
            # todo: Determine why this isn't set *always*

        if not msg.author:
            # sometimes discord will only send an author ID, not the author. this catches that
            msg.channel = (
                await self.cache.get_channel(to_snowflake(msg._channel_id)) if not msg.channel else msg.channel
            )
            if msg._guild_id:
                msg.guild = await self.cache.get_guild(msg._guild_id) if not msg.guild else msg.guild
                msg.author = await self.cache.get_member(msg._guild_id, msg._author_id)
            else:
                msg.author = await self.cache.get_user(to_snowflake(msg._author_id))

        self.dispatch(events.MessageCreate(msg))

    @listen()
    async def _on_raw_message_delete(self, event: RawGatewayEvent) -> None:
        """
        Process raw deletions and dispatch a processed deletion event.
        Args:
            event: raw message deletion event
        """
        message = await self.cache.get_message(event.data.get("channel_id"), event.data.get("id"))
        self.dispatch(events.MessageDelete(message))

    @listen()
    async def _on_raw_message_update(self, event: RawGatewayEvent) -> None:
        """
        Process raw message update event and dispatch a processed update event.

        Args:
            event: raw message update event
        """

        # a copy is made because the cache will update the original object in memory
        before = copy.copy(
            await self.cache.get_message(event.data.get("channel_id"), event.data.get("id"), request_fallback=False)
        )
        after = self.cache.place_message_data(event.data)
        self.dispatch(events.MessageUpdate(before=before, after=after))

    @listen()
    async def _on_raw_guild_create(self, event: RawGatewayEvent) -> None:
        """
        Automatically cache a guild upon GUILD_CREATE event from gateway.

        Args:
            data: raw guild data
        """
        guild = self.cache.place_guild_data(event.data)
        self._guild_event.set()

        self.dispatch(events.GuildCreate(guild))

    @listen()
    async def _on_raw_channel_create(self, event: RawGatewayEvent) -> None:
        """
        Automatically cache a guild upon CHANNEL_CREATE event from gateway.

        Args:
            data: raw channel data
        """
        channel = self.cache.place_channel_data(event.data)
        self.dispatch(events.ChannelCreate(channel))

    @listen()
    async def _on_raw_channel_delete(self, event: RawGatewayEvent) -> None:
        """
        Process raw channel deletions and dispatch a processed channel deletion event.

        Args:
            event: raw channel deletion event
        """
        # for some reason this event returns the deleted object?
        # so we cache it regardless
        channel = self.cache.place_channel_data(event.data)
        self.dispatch(events.ChannelDelete(channel))

    @listen()
    async def _on_raw_typing_start(self, event: RawGatewayEvent) -> None:
        """
        Process raw typing start and dispatch a processed typing event.

        Args:
            event: raw typing start event
        """
        author: Union[User, Member]
        channel: BaseChannel
        guild = None

        if member := event.data.get("member"):
            author = self.cache.place_member_data(event.data.get("guild_id"), member)
            guild = await self.cache.get_guild(event.data.get("guild_id"))
        else:
            author = await self.cache.get_user(event.data.get("user_id"))

        channel = await self.cache.get_channel(event.data.get("channel_id"))

        self.dispatch(
            events.TypingStart(
                author=author,
                channel=channel,
                guild=guild,
                timestamp=Timestamp.utcfromtimestamp(event.data.get("timestamp")),
            )
        )

    @listen()
    async def _on_websocket_ready(self, event: events.RawGatewayEvent) -> None:
        """
        Catches websocket ready and determines when to dispatch the client `READY` signal.

        Args:
            event: The websocket ready packet
        """
        data = event.data
        expected_guilds = set(to_snowflake(guild["id"]) for guild in data["guilds"])
        self._user._add_guilds(expected_guilds)

        while True:
            try:  # wait to let guilds cache
                await asyncio.wait_for(self._guild_event.wait(), self.guild_event_timeout)
            except asyncio.TimeoutError:
                log.warning("Timeout waiting for guilds cache: Not all guilds will be in cache")
                break
            self._guild_event.clear()

            if len(self.cache.guild_cache) == len(expected_guilds):
                # all guilds cached
                break

        # cache slash commands
        await self._init_interactions()

        self._ready = True
        self.dispatch(events.Ready())

    def get_scale(self, name) -> Optional[Scale]:
        """
        Get a scale
        Args:
            name: The name of the scale, or the name of it's extension

        Returns:
            Scale or None if no scale is found
        """
        if name not in self.scales.keys():
            for scale in self.scales.values():
                if scale.extension_name == name:
                    return scale

        return self.scales.get(name, None)

    def grow_scale(self, file_name: str, package: str = None) -> None:
        """
        A helper method to load a scale

        Args:
            file_name: The name of the file to load the scale from.
            package: The package this scale is in.
        """
        self.load_extension(file_name, package)

    def shed_scale(self, scale_name: str) -> None:
        """
        Helper method to unload a scale

        Args:
            scale_name: The name of the scale to unload.
        """
        if scale := self.get_scale(scale_name):
            return self.unload_extension(inspect.getmodule(scale).__name__)

        raise ScaleLoadException(f"Unable to shed scale: No scale exists with name: `{scale_name}`")

    def regrow_scale(self, scale_name: str) -> None:
        """
        Helper method to reload a scale.

        Args:
            scale_name: The name of the scale to reload
        """
        self.shed_scale(scale_name)
        self.grow_scale(scale_name)

    def load_extension(self, name: str, package: str = None):
        """
        Load an extension.

        Args:
            name: The name of the extension.
            package: The package the extension is in
        """
        name = importlib.util.resolve_name(name, package)
        if name in self.__extensions:
            raise Exception(f"{name} already loaded")

        module = importlib.import_module(name, package)
        try:
            setup = getattr(module, "setup")
            setup(self)
        except Exception as e:
            del sys.modules[name]
            raise ExtensionLoadException(f"Error loading {name}") from e

        else:
            log.debug(f"Loaded Extension: {name}")
            self.__extensions[name] = module
            return

    def unload_extension(self, name, package=None):
        """
        unload an extension.

        Args:
            name: The name of the extension.
            package: The package the extension is in
        """
        name = importlib.util.resolve_name(name, package)
        module = self.__extensions.get(name)

        if module is None:
            raise ExtensionNotFound(f"No extension called {name} is loaded")

        try:
            teardown = getattr(module, "teardown")
            teardown()
        except AttributeError:
            pass

        if scale := self.get_scale(name):
            scale.shed()

        del sys.modules[name]
        del self.__extensions[name]

    def reload_extension(self, name, package=None):
        """
        Helper method to reload an extension.
        Simply unloads, then loads the extension.

        Args:
            name: The name of the extension.
            package: The package the extension is in
        """
        name = importlib.util.resolve_name(name, package)
        module = self.__extensions.get(name)

        if module is None:
            log.warning("Attempted to reload extension thats not loaded. Loading extension instead")
            return self.load_extension(name, package)

        self.unload_extension(name, package)
        self.load_extension(name, package)

        # todo: maybe add an ability to revert to the previous version if unable to load the new one

    async def get_guild(self, guild_id: "Snowflake_Type") -> Guild:
        """
        Get a guild

        Note:
            This method is an alias for the cache which will either return a cached object, or query discord for the object
            if its not already cached.

        Args:
            guild_id: The ID of the guild to get

        Returns:
            Guild Object
        """
        return await self.cache.get_guild(guild_id)

    async def get_channel(self, channel_id: "Snowflake_Type") -> BaseChannel:
        """
        Get a channel

        Note:
            This method is an alias for the cache which will either return a cached object, or query discord for the object
            if its not already cached.

        Args:
            channel_id: The ID of the channel to get

        Returns:
            Channel Object
        """
        return await self.cache.get_channel(channel_id)

    async def get_user(self, user_id: "Snowflake_Type") -> User:
        """
        Get a user

        Note:
            This method is an alias for the cache which will either return a cached object, or query discord for the object
            if its not already cached.

        Args:
            user_id: The ID of the user to get

        Returns:
            User Object
        """
        return await self.cache.get_user(user_id)

    async def get_member(self, user_id: "Snowflake_Type", guild_id: "Snowflake_Type") -> Member:
        """
        Get a member from a guild

        Note:
            This method is an alias for the cache which will either return a cached object, or query discord for the object
            if its not already cached.

        Args:
            user_id: The ID of the member
            guild_id: The ID of the guild to get the member from

        Returns:
            Member object
        """
        return await self.cache.get_member(guild_id, user_id)

    async def get_sticker(self, sticker_id: "Snowflake_Type"):
        sticker_data = await self.http.get_sticker(sticker_id)
        return Sticker.from_dict(sticker_data, self)

    async def get_nitro_packs(self) -> List["StickerPack"]:
        packs_data = await self.http.list_nitro_sticker_packs()
        packs = []
        for pack_data in packs_data:
            packs.append(StickerPack.from_dict(pack_data, self))
        return packs

    async def change_presence(
        self, status: Optional[Union[str, Status]] = Status.ONLINE, activity: Optional[Union[Activity, str]] = None
    ):
        """
        Change the bots presence.

        Args:
            status: The status for the bot to be. i.e. online, afk, etc.
            activity: The activity for the bot to be displayed as doing.

        note::
            Bots may only be `playing` `streaming` or `listening`, other activity types are likely to fail.
        """
        if activity:
            if not isinstance(activity, Activity):
                # squash whatever the user passed into an activity
                activity = Activity.create(name=str(activity))

            if activity.type == ActivityType.STREAMING:
                if not activity.url:
                    log.warning("Streaming activity cannot be set without a valid URL attribute")
            elif activity.type not in [ActivityType.GAME, ActivityType.STREAMING, ActivityType.LISTENING]:
                log.warning(f"Activity type `{ActivityType(activity.type).name}` may not be enabled for bots")

        if status:
            if not isinstance(status, Status):
                try:
                    status = Status[status.upper()]
                except KeyError:
                    raise ValueError(f"`{status}` is not a valid status type. Please use the Status enum") from None
        else:
            # in case the user set status to None
            log.warning("Status must be set to a valid status type, defaulting to online")
            status = Status.ONLINE

        await self.ws.change_presence(activity.to_dict() if activity else None, status)



"""These are events dispatched by the client. This is intended as a reference so you know what data to expect for each event

!!! Hint "Example Usage:"
    The event classes outlined here are in `CamelCase` to comply with Class naming convention, however the event names
    are actually in `lower_case_with_underscores` so your listeners should be named as following:

    ```python
    @bot.event
    def on_ready():
        # ready events pass no data, so dont have params
        print("Im ready!")

    @bot.event
    def on_guild_create(event):
        # guild_create events pass a guild object, expect a single param
        print(f"{event.guild.name} created")
    ```
"""
import re
from typing import TYPE_CHECKING

import attr

if TYPE_CHECKING:
    from dis_snek.models.discord_objects.context import ComponentContext
    from dis_snek.models.snowflake import Snowflake_Type

_event_reg = re.compile("(?<!^)(?=[A-Z])")


@attr.s()
class BaseEvent:
    """A base event that all other events inherit from"""

    override_name: str = attr.ib(kw_only=True, default=None)

    @property
    def resolved_name(self):
        name = self.override_name or self.__class__.__name__
        return _event_reg.sub("_", name).lower()


@attr.s()
class GuildEvent:
    """A base event that adds guild_id"""

    guild_id: "Snowflake_Type" = attr.ib(metadata={"docs": "The ID of the guild"})


@attr.s(slots=True)
class Login(BaseEvent):
    """The bot has just logged in"""


@attr.s(slots=True)
class Connect(BaseEvent):
    """The bot is now connected to the discord Gateway."""


@attr.s(slots=True)
class Resume(BaseEvent):
    """The bot has resumed its connection to the discord Gateway"""


@attr.s(slots=True)
class Disconnect(BaseEvent):
    """The bot has just disconnected."""


@attr.s(slots=True)
class Ready(BaseEvent):
    """The client is now ready.

    Note:
        Don't use this event for things that must only happen once, on startup, as this event may be called multiple times.
    """


class WebsocketReady(BaseEvent):
    """The gateway has reported that it is ready"""


@attr.s(slots=True)
class Component(BaseEvent):
    """Dispatched when a user uses a Component"""

    context: "ComponentContext" = attr.ib(metadata={"docs": "The context of the interaction"})


@attr.s(slots=True)
class Button(Component):
    """Dispatched when a user uses a Button"""


@attr.s(slots=True)
class Select(Component):
    """Dispatched when a user uses a Select"""



from typing import TYPE_CHECKING, Any, List

from dis_snek.models.route import Route
from dis_snek.utils.serializer import dict_filter_none

if TYPE_CHECKING:
    from dis_snek.models.snowflake import Snowflake_Type


class WebhookRequests:
    request: Any

    async def create_webhook(self, channel_id: "Snowflake_Type", name: str, avatar: Any = None) -> dict:
        """
        Create a new webhook.

        parameters:
            channel_id: The id of the channel to add this webhook to
            name: name of the webhook (1-80 characters)
            avatar: The image for the default webhook avatar
        """
        return await self.request(
            Route("POST", f"/channels/{channel_id}/webhooks"), data={"name": name, "avatar": avatar}
        )

    async def get_channel_webhooks(self, channel_id: "Snowflake_Type") -> List[dict]:
        """
        Return a list of channel webhook objects.

        parameters:
            channel_id: The id of the channel to query
        returns:
            List of webhook objects
        """
        return await self.request(Route("GET", f"/channels/{channel_id}/webhooks"))

    async def get_guild_webhooks(self, guild_id: "Snowflake_Type") -> List[dict]:
        """
        Return a list of guild webhook objects.

        parameters:
            guild_id: The id of the guild to query
        returns:
            List of webhook objects
        """
        return await self.request(Route("GET", f"/guilds/{guild_id}/webhooks"))

    async def get_webhook(self, webhook_id: "Snowflake_Type", webhook_token: str = None) -> dict:
        """
        Return the new webhook object for the given id.

        parameters:
            webhook_id: The ID of the webhook to get
            webhook_token: The token for the webhook
        returns:
            Webhook object
        """
        endpoint = f"/webhooks/{webhook_id}{f'/{webhook_token}' if webhook_token else ''}"

        return await self.request(Route("GET", endpoint))

    async def modify_webhook(
        self,
        webhook_id: "Snowflake_Type",
        name: str,
        avatar: Any,
        channel_id: "Snowflake_Type",
        webhook_token: str = None,
    ) -> dict:
        """
        Modify a webhook.

        parameters:
            name: the default name of the webhook
            avatar: image for the default webhook avatar
            channel_id: the new channel id this webhook should be moved to
            webhook_id: The ID of the webhook to modify
            webhook_token: The token for the webhook

        """
        endpoint = f"/webhooks/{webhook_id}{f'/{webhook_token}' if webhook_token else ''}"

        return await self.request(
            Route("PATCH", endpoint), data={"name": name, "avatar": avatar, "channel_id": channel_id}
        )

    async def delete_webhook(self, webhook_id: "Snowflake_Type", webhook_token: str = None) -> dict:
        """
        Delete a webhook

        parameters:
            webhook_id: The ID of the webhook to delete
            webhook_token: The token for the webhook
        returns:
            Webhook object
        """
        endpoint = f"/webhooks/{webhook_id}{f'/{webhook_token}' if webhook_token else ''}"

        return await self.request(Route("DELETE", endpoint))

    async def execute_webhook(
        self,
        webhook_id: "Snowflake_Type",
        webhook_token: str,
        payload: dict,
        wait=False,
        thread_id: "Snowflake_Type" = None,
    ) -> dict:
        """
        Execute a webhook. Basically send a message as the webhook.

        parameters:
            webhook_id: The ID of the webhook to delete
            webhook_token: The token for the webhook
            payload: The JSON payload for the message
            wait: Waits for server confirmation of message send before response
            thread_id: Send a message to the specified thread
            suffix: An optional suffix to add to the end of the endpoint address
        returns:
            The sent `message`, if `wait` is True else None
        """
        return await self.request(
            Route("POST", f"/webhooks/{webhook_id}/{webhook_token}"),
            params=dict_filter_none({"wait": "true" if wait else "false", "thread_id": thread_id}),
            data=payload,
        )

    async def get_webhook_message(
        self, webhook_id: "Snowflake_Type", webhook_token: str, message_id: "Snowflake_Type"
    ) -> dict:
        """
        Returns a previously-sent webhook message from the same token. Returns a message object on success.

        parameters:
            webhook_id: The ID of the webhook to delete
            webhook_token: The token for the webhook
            message_id: The ID of a message sent by this webhook
        returns:
            A message object on success
        """
        return await self.request(Route("GET", f"/webhooks/{webhook_id}/{webhook_token}/messages/{message_id}"))

    async def edit_webhook_message(
        self, webhook_id: "Snowflake_Type", webhook_token: str, message_id: "Snowflake_Type", payload: dict
    ) -> dict:
        """
        Edits a previously-sent webhook message from the same token.

        parameters:
            webhook_id: The ID of the webhook to delete
            webhook_token: The token for the webhook
            message_id: The ID of a message sent by this webhook
            payload: The JSON payload for the message
        returns:
            The updated message on success
        """
        return await self.request(
            Route("PATCH", f"/webhooks/{webhook_id}/{webhook_token}/messages/{message_id}"), data=payload
        )

    async def delete_webhook_message(
        self, webhook_id: "Snowflake_Type", webhook_token: str, message_id: "Snowflake_Type"
    ) -> None:
        """
        Delete a message that was created by the same token.

        parameters:
            webhook_id: The ID of the webhook to delete
            webhook_token: The token for the webhook
            message_id: The ID of a message sent by this webhook
        """
        return await self.request(Route("DELETE", f"/webhooks/{webhook_id}/{webhook_token}/messages/{message_id}"))



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union


from dis_snek.const import MISSING
from dis_snek.models.route import Route
from dis_snek.utils.serializer import dict_filter_none, dict_filter_missing

if TYPE_CHECKING:
    from dis_snek.models.discord_objects.channel import PermissionOverwrite
    from dis_snek.models.enums import ChannelTypes
    from dis_snek.models.snowflake import Snowflake_Type


class ChannelRequests:
    request: Any

    async def get_channel(self, channel_id: "Snowflake_Type") -> dict:
        """
        Get a channel by ID. Returns a channel object. If the channel is a thread, a thread member object is included.

        parameters:
            channel_id: The id of the channel
        returns:
            channel
        """
        return await self.request(Route("GET", f"/channels/{channel_id}"))

    async def get_channel_messages(
        self,
        channel_id: "Snowflake_Type",
        limit: int = 50,
        around: Optional["Snowflake_Type"] = None,
        before: Optional["Snowflake_Type"] = None,
        after: Optional["Snowflake_Type"] = None,
    ) -> List[dict]:
        """
        Get the messages for a channel.

        parameters:
            channel_id: The channel to get messages from
            limit: How many messages to get (default 50, max 100)
            around: Get messages around this snowflake
            before: Get messages before this snowflake
            after: Get messages after this snowflake

        returns:
            List of message dicts
        """
        params: Dict[str, Union[int, str]] = {"limit": limit}

        params_used = 0

        if before:
            params_used += 1
            params["before"] = before
        if after:
            params_used += 1
            params["after"] = after
        if around:
            params_used += 1
            params["around"] = around

        if params_used > 1:
            raise ValueError("`before` `after` and `around` are mutually exclusive, only one may be passed at a time.")

        return await self.request(Route("GET", f"/channels/{channel_id}/messages"), params=params)

    async def create_guild_channel(
        self,
        guild_id: "Snowflake_Type",
        name: str,
        channel_type: Union["ChannelTypes", int],
        topic: Optional[str] = MISSING,
        position: int = 0,
        permission_overwrites: List[Union["PermissionOverwrite", dict]] = MISSING,
        parent_id: "Snowflake_Type" = MISSING,
        nsfw: bool = False,
        bitrate: int = 64000,
        user_limit: int = 0,
        rate_limit_per_user: int = 0,
        reason: str = MISSING,
    ) -> Dict:
        """"""
        payload = dict(
            name=name,
            type=channel_type,
            topic=topic,
            position=position,
            rate_limit_per_user=rate_limit_per_user,
            nsfw=nsfw,
            parent_id=parent_id,
            permission_overwrites=permission_overwrites,
        )

        if channel_type in (2, 13):
            payload.update(
                bitrate=bitrate,
                user_limit=user_limit,
            )

        # clean up payload
        payload = dict_filter_missing(payload)
        return await self.request(Route("POST", f"/guilds/{guild_id}/channels"), data=payload, reason=reason)

    async def move_channel(
        self,
        guild_id: "Snowflake_Type",
        channel_id: "Snowflake_Type",
        new_pos: int,
        parent_id: "Snowflake_Type" = None,
        lock_perms: bool = False,
        reason: str = MISSING,
    ) -> None:
        """
        Move a channel.

        parameters:
            guild_id: The ID of the guild this affects
            channel_id: The ID of the channel to move
            new_pos: The new position of this channel
            parent_id: The parent ID if needed
            lock_perms: Sync permissions with the new parent
            reason: An optional reason for the audit log
        """
        payload = dict(id=channel_id, position=new_pos, lock_permissions=lock_perms)
        if parent_id:
            payload["parent_id"] = parent_id

        return await self.request(Route("PATCH", f"/guilds/{guild_id}/channels"), data=payload, reason=reason)

    async def modify_channel(self, channel_id: "Snowflake_Type", data: dict, reason: str = MISSING) -> dict:
        """
        Update a channel's settings, returns the updated channel object on success.

        parameters:
            channel_id: The ID of the channel to update
            data: The data to update with
            reason: An optional reason for the audit log

        returns:
            Channel object on success
        """
        return await self.request(Route("PATCH", f"/channels/{channel_id}"), data=data, reason=reason)

    async def delete_channel(self, channel_id: "Snowflake_Type", reason: str = MISSING):
        """
        Delete the channel

        parameters:
            channel_id: The ID of the channel to delete
            reason: An optional reason for the audit log
        """
        return await self.request(Route("DELETE", f"/channels/{channel_id}"), reason=reason)

    async def get_channel_invites(self, channel_id: "Snowflake_Type") -> List[dict]:
        """
        Get the invites for the channel.

        parameters:
            channel_id: The ID of the channel to retrieve from

        returns:
            List of invite objects
        """
        return await self.request(Route("GET", f"/channels/{channel_id}/invites"))

    async def create_channel_invite(
        self,
        channel_id: "Snowflake_Type",
        max_age: int = 86400,
        max_uses: int = 0,
        temporary: bool = False,
        unique: bool = False,
        target_type: int = None,
        target_user_id: "Snowflake_Type" = None,
        target_application_id: "Snowflake_Type" = None,
        reason: str = MISSING,
    ) -> dict:
        """
        Create an invite for the given channel.

        parameters:
            channel_id: The ID of the channel to create an invite for
            max_age: duration of invite in seconds before expiry, or 0 for never. between 0 and 604800 (7 days) (default 24 hours)
            max_uses: max number of uses or 0 for unlimited. between 0 and 100
            temporary: whether this invite only grants temporary membership
            unique: if true, don't try to reuse a similar invite (useful for creating many unique one time use invites)
            target_type: the type of target for this voice channel invite
            target_user_id: the id of the user whose stream to display for this invite, required if target_type is 1, the user must be streaming in the channel
            target_application_id: the id of the embedded application to open for this invite, required if target_type is 2, the application must have the EMBEDDED flag
            reason: An optional reason for the audit log

        returns:
            an invite object
        """
        payload = dict(max_age=max_age, max_uses=max_uses, temporary=temporary, unique=unique)
        if target_type:
            payload["target_type"] = target_type
        if target_user_id:
            payload["target_user_id"] = target_user_id
        if target_application_id:
            payload["target_application_id"] = target_application_id

        return await self.request(Route("POST", f"/channels/{channel_id}/invites"), data=payload, reason=reason)

    async def delete_invite(self, invite_code: str, reason: str = MISSING) -> dict:
        """
        Delete an invite.


        parameters:
            invite_code: The code of the invite to delete
            reason: The reason to delete the invite

        returns:
            The deleted invite object
        """
        return await self.request(Route("DELETE", f"/invites/{invite_code}"))

    async def edit_channel_permission(
        self,
        channel_id: "Snowflake_Type",
        overwrite_id: "Snowflake_Type",
        allow: str,
        deny: str,
        perm_type: int,
        reason: str = MISSING,
    ) -> None:
        """
        Edit the channel permission overwrites for a user or role in a channel.

        parameters:
            channel_id: The id of the channel
            overwrite_id: The id of the object to override
            allow: the bitwise value of all allowed permissions
            deny: the bitwise value of all disallowed permissions
            perm_type: 0 for a role or 1 for a member
            reason: The reason for this action
        """
        return await self.request(
            Route("PUT", f"/channels/{channel_id}/permissions/{overwrite_id}"),
            data={"allow": allow, "deny": deny, "type": perm_type},
            reason=reason,
        )

    async def delete_channel_permission(
        self, channel_id: "Snowflake_Type", overwrite_id: int, reason: str = MISSING
    ) -> None:
        """
        Delete a channel permission overwrite for a user or role in a channel.

        parameters:
            channel_id: The ID of the channel.
            overwrite_id: The ID of the overwrite
            reason: An optional reason for the audit log
        """
        return await self.request(Route("DELETE", f"/channels/{channel_id}/{overwrite_id}"), reason=reason)

    async def follow_news_channel(self, channel_id: "Snowflake_Type", webhook_channel_id: "Snowflake_Type") -> dict:
        """
        Follow a news channel to send messages to the target channel.

        parameters:
            channel_id: The channel to follow
            webhook_channel_id: ID of the target channel

        returns:
            Followed channel object
        """
        return await self.request(
            Route("POST", f"/channels/{channel_id}/followers"), data={"webhook_channel_id": webhook_channel_id}
        )

    async def trigger_typing_indicator(self, channel_id: "Snowflake_Type") -> None:
        """
        Post a typing indicator for the specified channel. Generally bots should not implement this route.

        parameters:
            channel_id: The id of the channel to "type" in
        """
        return await self.request(Route("POST", f"/channels/{channel_id}/typing"))

    async def get_pinned_messages(self, channel_id: "Snowflake_Type") -> List[dict]:
        """
        Get all pinned messages from a channel.

        parameters:
            channel_id: The ID of the channel to get pins from

        returns:
            A list of pinned message objects
        """
        return await self.request(Route("GET", f"/channels/{channel_id}/pins"))

    async def create_stage_instance(
        self, channel_id: "Snowflake_Type", topic: str, privacy_level: int = 1, reason: str = MISSING
    ) -> dict:
        """
        Create a new stage instance.

        parameters:
            channel_id: The ID of the stage channel
            topic: The topic of the stage instance (1-120 characters)
            privacy_level: Them privacy_level of the stage instance (default guild only)
            reason: The reason for the creating the stage instance

        returns:
            The stage instance
        """
        # todo: convert privacy_level to int-enum
        return await self.request(
            Route("POST", "/stage-instances"),
            data={
                "channel_id": channel_id,
                "topic": topic,
                "privacy_level": privacy_level,
            },
            reason=reason,
        )

    async def get_stage_instance(self, channel_id: "Snowflake_Type") -> dict:
        """
        Get the stage instance associated with a given channel, if it exists.

        parameters:
            channel_id: The ID of the channel to retrieve the instance for.

        returns:
            A stage instance.
        """
        return await self.request(Route("GET", f"/stage-instances/{channel_id}"))

    async def modify_stage_instance(
        self, channel_id: "Snowflake_Type", topic: str = None, privacy_level: int = None, reason: str = MISSING
    ) -> dict:
        """
        Update the fields of a given stage instance.

        parameters:
            channel_id: The id of the stage channel.
            topic: The new topic for the stage instance
            privacy_level: The privacy level for the stage instance
            reason: The reason for the change

        returns:
            The updated stage instance.
        """
        return await self.request(
            Route("PATCH", f"/stage-instances/{channel_id}"),
            data=dict_filter_none({"topic": topic, "privacy_level": privacy_level}),
            reason=reason,
        )

    async def delete_stage_instance(self, channel_id: "Snowflake_Type", reason: str = MISSING) -> None:
        """
        Delete a stage instance.

        parameters:
            channel_id: The ID of the channel to delete the stage instance for.
            reason: The reason for the deletion
        """
        return await self.request(Route("DELETE", f"/stage-instances/{channel_id}"), reason=reason)



from typing import TYPE_CHECKING, List, Optional

import attr

from dis_snek.models.discord import DiscordObject
from dis_snek.models.enums import TeamMembershipState
from dis_snek.utils.attr_utils import define

if TYPE_CHECKING:
    from dis_snek.models.discord_objects.user import User
    from dis_snek.models.snowflake import Snowflake_Type


@define()
class TeamMember:
    membership_state: TeamMembershipState = attr.ib(converter=TeamMembershipState)
    permissions: List[str] = attr.ib(default=["*"])
    team_id: "Snowflake_Type" = attr.ib()
    user: "User" = attr.ib()  # TODO: cache partial user (avatar, discrim, id, username)


@define()
class Team(DiscordObject):
    icon: Optional[str] = attr.ib(default=None)
    members: List[TeamMember] = attr.ib(factory=list)
    name: str = attr.ib()
    owner_user_id: "Snowflake_Type" = attr.ib()

    # TODO: Lots of functions related to getting items



import logging
from functools import partial
from typing import Any

import attr
from dis_snek.const import logger_name, MISSING

log = logging.getLogger(logger_name)


class_defaults = dict(
    eq=False,
    order=False,
    hash=False,
    slots=True,
    kw_only=True,
    on_setattr=[attr.setters.convert, attr.setters.validate],
)
field_defaults = dict(repr=False)


define = partial(attr.define, **class_defaults)
field = partial(attr.field, **field_defaults)


def copy_converter(value):
    if isinstance(value, (list, set)):
        return value.copy()
    return value


# def converter(attribute):
#     def decorator(func):
#         attribute.converter = func
#         return staticmethod(func)
#
#     return decorator


def str_validator(self, attribute: attr.Attribute, value: Any):
    if not isinstance(value, str):
        if value is MISSING:
            return
        setattr(self, attribute.name, str(value))
        log.warning(
            f"Value of {attribute.name} has been automatically converted to a string. Please use strings in future.\n"
            "Note: Discord will always return value as a string"
        )



"""
This file handles the interaction with discords http endpoints.
"""
import asyncio
import datetime
import logging
import traceback
from collections import defaultdict
from typing import Any, Dict, Optional, TypeVar, Union
from urllib.parse import quote as _uriquote

import aiohttp  # type: ignore
from aiohttp import BaseConnector, ClientResponse, ClientSession, ClientWebSocketResponse, FormData
from multidict import CIMultiDictProxy  # type: ignore

from dis_snek.const import __py_version__, __repo_url__, __version__, logger_name, MISSING
from dis_snek.errors import DiscordError, Forbidden, GatewayNotFound, HTTPException, NotFound, LoginError
from dis_snek.http_requests import (
    BotRequests,
    ChannelRequests,
    EmojiRequests,
    GuildRequests,
    InteractionRequests,
    MemberRequests,
    MessageRequests,
    ReactionRequests,
    StickerRequests,
    ThreadRequests,
    UserRequests,
    WebhookRequests,
)
from dis_snek.models.route import Route
from dis_snek.utils.input_utils import response_decode
from dis_snek.utils.serializer import dict_filter_missing

log = logging.getLogger(logger_name)


T = TypeVar("T")


class DiscordClientWebSocketResponse(ClientWebSocketResponse):
    """Represents the websocket connection with discord."""

    async def close(self, *, code: int = 4000, message: bytes = b"") -> bool:
        """
        Close the connection.

            code: The close code to use
            message: A message to send within the close
        """
        return await super().close(code=code, message=message)


class HTTPClient(
    BotRequests,
    ChannelRequests,
    EmojiRequests,
    GuildRequests,
    InteractionRequests,
    MemberRequests,
    MessageRequests,
    ReactionRequests,
    StickerRequests,
    ThreadRequests,
    UserRequests,
    WebhookRequests,
):
    """A http client for sending requests to the Discord API."""

    def __init__(self, connector: Optional[BaseConnector] = None, loop: Optional[asyncio.AbstractEventLoop] = None):
        self.connector: Optional[BaseConnector] = connector
        self.loop = asyncio.get_event_loop() if loop is None else loop
        self.__session: Optional[ClientSession] = MISSING
        self._retries: int = 5
        self.token: Optional[str] = None
        self.ratelimit_locks: Dict[str, asyncio.Lock] = defaultdict(asyncio.Lock)

        self.user_agent: str = (
            f"DiscordBot ({__repo_url__} {__version__} Python/{__py_version__}) aiohttp/{aiohttp.__version__}"
        )

    def __del__(self):
        if self.__session and not self.__session.closed:
            self.loop.run_until_complete(self.__session.close())

    @staticmethod
    def _parse_ratelimit(header: CIMultiDictProxy[str]) -> dict:
        """
        Parse the ratelimit data into a more usable format.

        parameters:
            header: the header of the response
        :return:
        """
        return {
            "bucket": header.get("x-ratelimit-bucket"),
            "limit": int(header.get("x-ratelimit-limit") or -1),
            "remaining": int(header.get("x-ratelimit-remaining") or -1),
            "delta": float(header.get("x-ratelimit-reset-after", 0)),  # type: ignore
            "time": datetime.datetime.utcfromtimestamp(float(header.get("x-ratelimit-reset", 0))),  # type: ignore
        }

    async def request(
        self, route: Route, data: Union[dict, FormData] = MISSING, reason: str = MISSING, **kwargs: Dict[str, Any]
    ) -> Any:
        """
        Make a request to discord

        parameters:
            route: The route to take
            json: A json payload to send in the request
            reason: Attach a reason to this request, used for audit logs
        """
        # Assemble headers
        headers: Dict[str, str] = {"User-Agent": self.user_agent}
        if self.token not in (None, MISSING):
            headers["Authorization"] = "Bot " + self.token

        if isinstance(data, list):
            headers["Content-Type"] = "application/json"
            kwargs["json"] = [dict_filter_missing(x) if isinstance(x, dict) else x for x in data]
        elif isinstance(data, dict):
            headers["Content-Type"] = "application/json"
            kwargs["json"] = dict_filter_missing(data)
        elif isinstance(data, FormData):
            kwargs["data"] = data

        if reason not in (None, MISSING):
            headers["X-Audit-Log-Reason"] = _uriquote(reason, safe="/ ")

        kwargs["headers"] = headers

        if route.rl_bucket is not None:
            lock = self.ratelimit_locks[route.rl_bucket]
        else:
            lock = asyncio.Lock()

        response: Optional[ClientResponse] = None
        result: Optional[Union[Dict[str, Any], str]] = None

        await lock.acquire()
        for tries in range(self._retries):
            try:
                async with self.__session.request(route.method, route.url, **kwargs) as response:
                    result = await response_decode(response)
                    r_limit_data = self._parse_ratelimit(response.headers)

                    if response.status == 429:
                        # ratelimit exceeded
                        log.error(
                            f"{route.method}::{route.url}: Has exceeded ratelimit! Reset in {r_limit_data['delta']} seconds"
                        )
                        await asyncio.sleep(r_limit_data["delta"])
                        continue

                    elif r_limit_data["remaining"] == 0:
                        # ratelimit about to be exceeded, stop calls
                        log.debug(
                            f"{route.method}::{route.url}: Has exhausted its ratelimit! Locking route for {r_limit_data['delta']} seconds"
                        )
                        self.loop.call_later(r_limit_data["delta"], lock.release)

                    elif response.status in {500, 502, 504}:
                        # server issues, retry
                        log.warning(
                            f"{route.method}::{route.url}: Received {response.status}... retrying in {1 + tries * 2} seconds"
                        )
                        await asyncio.sleep(1 + tries * 2)
                        continue

                    if not 300 > response.status >= 200:
                        if not r_limit_data["remaining"] == 0:
                            lock.release()
                        await self._raise_exception(response, route, result)

                    # Success!
                    log.debug(f"{route.method}::{route.url}: Received {response.status}, releasing")
                    if not r_limit_data["remaining"] == 0:
                        lock.release()
                    return result

            except OSError as e:
                if tries < self._retries - 1 and e.errno in (54, 10054):
                    await asyncio.sleep(1 + tries * 2)
                    continue
                try:
                    lock.release()
                except RuntimeError:
                    pass
                raise
            except (Forbidden, NotFound, DiscordError, HTTPException):
                raise
            except Exception as e:
                try:
                    lock.release()
                except RuntimeError:
                    pass
                log.error("".join(traceback.format_exception(type(e), e, e.__traceback__)))
                break
        if lock.locked():
            # be clean and make sure we unlock
            lock.release()

    async def _raise_exception(self, response, route, result):

        if response.status == 403:
            raise Forbidden(response, response_data=result, route=route)
        elif response.status == 404:
            raise NotFound(response, response_data=result, route=route)
        elif response.status >= 500:
            raise DiscordError(response, response_data=result, route=route)
        else:
            raise HTTPException(response, response_data=result, route=route)

    async def request_cdn(self, url, asset) -> bytes:
        log.debug(f"{asset} requests {url} from CDN")
        async with self.__session.get(url) as response:
            if response.status == 200:
                return await response.read()
            await self._raise_exception(response, asset, await response_decode(response))

    async def login(self, token: str) -> dict:
        """
        "Login" to the gateway, basically validates the token and grabs user data.

        parameters:
            token: the token to use
        returns:
            The currently logged in bot's data
        """
        self.__session = ClientSession(connector=self.connector, ws_response_class=DiscordClientWebSocketResponse)
        self.token = token
        try:
            return await self.request(Route("GET", "/users/@me"))
        except HTTPException as e:
            if e.status == 401:
                raise LoginError("An improper token was passed") from e
            raise

    async def close(self) -> None:
        """Close the session."""
        if self.__session:
            await self.__session.close()

    async def logout(self) -> None:
        """Logout of the session."""
        await self.request(Route("POST", "/auth/logout"))

    async def get_gateway(self) -> str:
        """Get the gateway url."""
        try:
            data: dict = await self.request(Route("GET", "/gateway"))
        except HTTPException as exc:
            raise GatewayNotFound from exc
        return "{0}?encoding={1}&v=9&compress=zlib-stream".format(data["url"], "json")

    async def websocket_connect(self, url: str) -> ClientWebSocketResponse:
        """
        Connect to the websocket.

        parameters:
            url: the url to connect to
        """
        return await self.__session.ws_connect(
            url, timeout=30, max_msg_size=0, autoclose=False, headers={"User-Agent": self.user_agent}, compress=0
        )



from typing import Optional, List

import attr

from dis_snek.models.discord import ClientObject
from dis_snek.models.discord_objects.asset import Asset
from dis_snek.models.enums import ActivityType
from dis_snek.models.snowflake import Snowflake_Type
from dis_snek.models.timestamp import Timestamp
from dis_snek.utils.attr_utils import define, field, class_defaults
from dis_snek.utils.serializer import dict_filter_none


@define()
class Activity:
    name: str = attr.ib()
    type: ActivityType = attr.ib(default=ActivityType.GAME)
    url: Optional[str] = attr.ib(default=None)
    created_at: Timestamp = attr.ib(default=None)
    timestamps: Timestamp = attr.ib(default=None)
    application_id: "Snowflake_Type" = attr.ib(default=None)
    details: Optional[str] = attr.ib(default=None)
    state: Optional[str] = attr.ib(default=None)
    emoji: Optional[str] = attr.ib(default=None)
    party: Optional[dict] = attr.ib(default=None)  # todo: Create party object
    assets: Optional[Asset] = attr.ib(default=None)
    secrets: Optional[str] = attr.ib(default=None)
    instance: Optional[bool] = attr.ib(default=False)
    flags: Optional[int] = attr.ib(default=None)  # todo: Activity Flags
    buttons: List[dict] = attr.ib(factory=list)  # todo: Activity Button Object

    @classmethod
    def create(cls, name: str, type: ActivityType = ActivityType.GAME, url: Optional[str] = None) -> "Activity":
        return cls(name=name, type=type, url=url)  # noqa

    def to_dict(self) -> dict:
        return dict_filter_none({"name": self.name, "type": self.type, "url": self.url})

    # todo: handle incoming presence data



import asyncio
import logging
from typing import Callable, Coroutine, Any

import attr

from dis_snek.const import MISSING, logger_name
from dis_snek.errors import CommandOnCooldown, CommandCheckFailure, MaxConcurrencyReached
from dis_snek.mixins.serialization import DictSerializationMixin
from dis_snek.models.cooldowns import Cooldown, Buckets, MaxConcurrency
from dis_snek.utils.serializer import no_export_meta

log = logging.getLogger(logger_name)


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class BaseCommand(DictSerializationMixin):
    """
    An object all commands inherit from.
    Outlines the basic structure of a command, and handles checks.

    attributes:
        scale: The scale this command belongs to.
        enabled: Whether this command is enabled
        checks: Any checks that must be run before this command can be run
        callback: The coroutine to be called for this command
        error_callback: The coroutine to be called when an error occurs
        pre_run_callback: A coroutine to be called before this command is run **but** after the checks
        post_run_callback: A coroutine to be called after this command has run

    """

    scale: Any = attr.ib(default=None, metadata=no_export_meta)
    """The scale this command belongs to"""
    enabled: bool = attr.ib(default=True, metadata=no_export_meta)
    """Whether this can be run at all"""

    checks: list = attr.ib(factory=list)
    """Any checks that must be *checked* before the command can run"""
    cooldown: Cooldown = attr.ib(default=MISSING)
    """An optional cooldown to apply to the command"""
    max_concurrency: MaxConcurrency = attr.ib(default=MISSING)
    """An optional maximum number of concurrent instances to apply to the command"""

    callback: Callable[..., Coroutine] = attr.ib(default=None, metadata=no_export_meta)
    """The coroutine to be called for this command"""
    error_callback: Callable[..., Coroutine] = attr.ib(default=None, metadata=no_export_meta)
    """The coroutine to be called when an error occurs"""
    pre_run_callback: Callable[..., Coroutine] = attr.ib(default=None, metadata=no_export_meta)
    """The coroutine to be called before the command is executed, **but** after the checks"""
    post_run_callback: Callable[..., Coroutine] = attr.ib(default=None, metadata=no_export_meta)
    """The coroutine to be called after the command has executed"""

    def __attrs_post_init__(self):
        if self.callback is not None:
            if hasattr(self.callback, "checks"):
                self.checks += self.callback.checks
            if hasattr(self.callback, "cooldown"):
                self.cooldown = self.callback.cooldown
            if hasattr(self.callback, "max_concurrency"):
                self.max_concurrency = self.callback.max_concurrency

    async def __call__(self, context, *args, **kwargs):
        """
        Calls this command.

        parameters:
            context: The context of this command
            args: Any
            kwargs: Any
        """
        try:
            if await self._can_run(context):
                if self.pre_run_callback is not None:
                    await self.pre_run_callback(context, *args, **kwargs)

                if self.scale is not None and self.scale.scale_prerun:
                    await self.scale.scale_prerun(context, *args, **kwargs)

                await self.callback(context, *args, **kwargs)

                if self.post_run_callback is not None:
                    await self.post_run_callback(context, *args, **kwargs)

                if self.scale is not None and self.scale.scale_postrun:
                    await self.scale.scale_postrun(context, *args, **kwargs)

        except Exception as e:
            if self.error_callback:
                await self.error_callback(e, context, *args, **kwargs)
            else:
                raise
        finally:
            if self.max_concurrency is not MISSING:
                await self.max_concurrency.release(context)

    async def _can_run(self, context):
        """
        Determines if this command can be run.

        parameters:
            context: The context of the command
        """
        max_conc_acquired = False  # signals if a semaphore has been acquired, for exception handling

        try:
            if not self.enabled:
                return False

            for _c in self.checks:
                if not await _c(context):
                    raise CommandCheckFailure(self, _c)

            if self.scale and self.scale.scale_checks:
                for _c in self.scale.scale_checks:
                    if not await _c(context):
                        raise CommandCheckFailure(self, _c)

            if self.max_concurrency is not MISSING:
                if not await self.max_concurrency.acquire(context):
                    raise MaxConcurrencyReached(self, self.max_concurrency)

            if self.cooldown is not MISSING:
                if not await self.cooldown.acquire_token(context):
                    raise CommandOnCooldown(self, await self.cooldown.get_cooldown(context))

            return True

        except Exception:
            if max_conc_acquired:
                await self.max_concurrency.release(context)
            raise

    def error(self, call: Callable[..., Coroutine]):
        """A decorator to declare a coroutine as one that will be run upon an error."""
        if not asyncio.iscoroutinefunction(call):
            raise TypeError("Error handler must be coroutine")
        self.error_callback = call
        return call

    def pre_run(self, call: Callable[..., Coroutine]):
        """A decorator to declare a coroutine as one that will be run before the command"""
        if not asyncio.iscoroutinefunction(call):
            raise TypeError("pre_run must be coroutine")
        self.pre_run_callback = call
        return call

    def post_run(self, call: Callable[..., Coroutine]):
        """A decorator to declare a coroutine as one that will be run after the command has"""
        if not asyncio.iscoroutinefunction(call):
            raise TypeError("post_run must be coroutine")
        self.post_run_callback = call
        return call


@attr.s(slots=True, kw_only=True, on_setattr=[attr.setters.convert, attr.setters.validate])
class MessageCommand(BaseCommand):
    """
    Represents a command triggered by standard message.
    """

    name: str = attr.ib()
    """The name of the command."""


def message_command(
    name: str = None,
):
    """
    A decorator to declare a coroutine as a message command.

    parameters:
        name: The name of the command, defaults to the name of the coroutine
    returns:
        Message Command Object
    """

    def wrapper(func):
        if not asyncio.iscoroutinefunction(func):
            raise ValueError("Commands must be coroutines")
        cmd = MessageCommand(name=name or func.__name__, callback=func)
        return cmd

    return wrapper


def check(check: Callable[..., Coroutine]):
    """
    Add a check to a command.

    parameters:
        check: A coroutine as a check for this command
    """

    def wrapper(coro):
        if isinstance(coro, BaseCommand):
            coro.checks.append(check)
            return
        if not hasattr(coro, "checks"):
            coro.checks = []
        coro.checks.append(check)
        return coro

    return wrapper


def cooldown(bucket: Buckets, rate: int, interval: float):
    """
    Add a cooldown to a command

    Args:
        bucket: The bucket used to track cooldowns
        rate: How many commands may be ran per interval
        interval: How many seconds to wait for a cooldown
    """

    def wrapper(coro: Callable[..., Coroutine]):
        cooldown_obj = Cooldown(bucket, rate, interval)

        coro.cooldown = cooldown_obj

        return coro

    return wrapper


def max_concurrency(bucket: Buckets, concurrent: int):
    """
    Add a maximum number of concurrent instances to the command.

    Args:
        bucket: The bucket to enforce the maximum within
        concurrent: The maximum number of concurrent instances to allow
    """

    def wrapper(coro: Callable[..., Coroutine]):
        max_conc = MaxConcurrency(concurrent, bucket)

        coro.max_concurrency = max_conc

        return coro

    return wrapper



from enum import IntEnum
from typing import TYPE_CHECKING, List, Optional, Union

import attr
from attr.converters import optional

from dis_snek.const import MISSING
from dis_snek.models.discord import DiscordObject
from dis_snek.models.snowflake import to_snowflake
from dis_snek.utils.attr_utils import define
from dis_snek.utils.serializer import dict_filter_none

if TYPE_CHECKING:
    from dis_snek.models.discord_objects.guild import Guild
    from dis_snek.models.discord_objects.user import User
    from dis_snek.models.snowflake import Snowflake_Type


class StickerTypes(IntEnum):
    """Types of sticker."""

    STANDARD = 1
    """An official sticker in a pack, part of Nitro or in a removed purchasable pack."""
    GUILD = 2
    """A sticker uploaded to a Boosted guild for the guild's members."""


class StickerFormatTypes(IntEnum):
    """File formats for stickers."""

    PNG = 1
    APNG = 2
    LOTTIE = 3


@define(kw_only=False)
class StickerItem(DiscordObject):
    name: str = attr.ib()
    """Name of the sticker."""
    format_type: StickerFormatTypes = attr.ib(converter=StickerFormatTypes)
    """Type of sticker image format."""


@define()
class Sticker(StickerItem):
    """Represents a sticker that can be sent in messages."""

    pack_id: Optional["Snowflake_Type"] = attr.ib(default=None, converter=optional(to_snowflake))
    """For standard stickers, id of the pack the sticker is from."""
    description: Optional[str] = attr.ib(default=None)
    """Description of the sticker."""
    tags: str = attr.ib()
    """autocomplete/suggestion tags for the sticker (max 200 characters)"""
    type: Union[StickerTypes, int] = attr.ib(converter=StickerTypes)
    """Type of sticker."""
    available: Optional[bool] = attr.ib(default=True)
    """Whether this guild sticker can be used, may be false due to loss of Server Boosts."""
    sort_value: Optional[int] = attr.ib(default=None)
    """The standard sticker's sort order within its pack."""

    _user_id: Optional["Snowflake_Type"] = attr.ib(default=None, converter=optional(to_snowflake))
    _guild_id: Optional["Snowflake_Type"] = attr.ib(default=None, converter=optional(to_snowflake))

    async def get_creator(self) -> "User":
        """
        Get the user who created this emoji

        Returns:
            User object
        """
        return await self._client.cache.get_user(self._user_id)

    async def get_guild(self) -> "Guild":
        """
        Get the guild associated with this emoji

        Returns:
            Guild object
        """
        return await self._client.cache.get_guild(self._guild_id)

    async def edit(
        self,
        name: Optional[str] = MISSING,
        description: Optional[str] = MISSING,
        tags: Optional[str] = MISSING,
        reason: Optional[str] = MISSING,
    ) -> "Sticker":
        """
        # TODO
        """
        if not self._guild_id:
            raise ValueError("You can only edit guild stickers.")

        payload = dict_filter_none(dict(name=name, description=description, tags=tags))
        sticker_data = await self._client.http.modify_guild_sticker(payload, self._guild_id, self.id, reason)
        return self.update_from_dict(sticker_data)

    async def delete(self, reason: Optional[str] = MISSING):
        """
        # TODO
        """
        if not self._guild_id:
            raise ValueError("You can only delete guild stickers.")

        await self._client.http.delete_guild_sticker(self._guild_id, self.id, reason)


@define()
class StickerPack(DiscordObject):
    """Represents a pack of standard stickers."""

    stickers: List["Sticker"] = attr.ib(factory=list)
    """The stickers in the pack."""
    name: str = attr.ib()
    """Name of the sticker pack."""
    sku_id: "Snowflake_Type" = attr.ib()
    """id of the pack's SKU."""
    cover_sticker_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    """id of a sticker in the pack which is shown as the pack's icon."""
    description: str = attr.ib()
    """Description of the sticker pack."""
    banner_asset_id: "Snowflake_Type" = attr.ib()  # TODO CDN Asset
    """id of the sticker pack's banner image."""



from contextlib import suppress
from typing import Union

from dis_snek.models.timestamp import Timestamp
from dis_snek.utils.attr_utils import define, field

Snowflake_Type = Union[str, int]


def to_snowflake(snowflake: Union[Snowflake_Type, "SnowflakeObject"]) -> int:
    if isinstance(snowflake, SnowflakeObject):
        snowflake = snowflake.id

    if not isinstance(snowflake, (int, str)):
        raise TypeError(
            f"ID (snowflake) should be instance of int, str or SnowflakeObject. Got '{snowflake}' ({type(snowflake)}) "
            f"instead."
        )

    with suppress(ValueError):
        snowflake = int(snowflake)

    if 22 > snowflake.bit_length() > 64:
        raise ValueError("ID (snowflake) is not in correct discord format!")

    return snowflake


@define()
class SnowflakeObject:
    id: "Snowflake_Type" = field(repr=True, converter=to_snowflake)

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __hash__(self):
        return int(self.id) << 32

    @property
    def created_at(self) -> "Timestamp":
        """
        Returns a timestamp representing the date-time this discord object was created
        :return:
        """
        return Timestamp.from_snowflake(self.id)



from datetime import datetime
from typing import Union

from dis_snek.models.timestamp import Timestamp


def timestamp_converter(value: Union[datetime, int, float, str]) -> Timestamp:
    if isinstance(value, str):
        return Timestamp.fromisoformat(value)
    elif isinstance(value, (float, int)):
        return Timestamp.fromtimestamp(float(value))
    elif isinstance(value, datetime):
        return Timestamp.fromdatetime(value)
    raise TypeError("Timestamp must be one of: datetime, int, float, ISO8601 str")


def list_converter(converter):
    def convert_action(value):
        return [
            converter(element) if not hasattr(element, "to_dict") else converter(**element.to_dict())
            for element in value
        ]

    return convert_action



from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import attr

from dis_snek.const import MISSING
from dis_snek.models.color import Color
from dis_snek.models.discord import DiscordObject
from dis_snek.models.enums import Permissions
from dis_snek.utils.attr_utils import define, field
from dis_snek.utils.serializer import dict_filter_missing

if TYPE_CHECKING:
    from dis_snek.client import Snake
    from dis_snek.models.discord_objects.guild import Guild
    from dis_snek.models.discord_objects.user import Member
    from dis_snek.models.snowflake import Snowflake_Type


def sentinel_converter(value, sentinel=attr.NOTHING):
    if value is sentinel:
        return False
    elif value is None:
        return True
    return value


@define()
class Role(DiscordObject):
    _sentinel = object()

    name: str = field(repr=True)
    color: "Color" = field(converter=Color)
    hoist: bool = field(default=False)
    position: int = field(repr=True)
    permissions: "Permissions" = field(converter=Permissions)
    managed: bool = field(default=False)
    mentionable: bool = field(default=True)
    premium_subscriber: bool = field(default=_sentinel, converter=partial(sentinel_converter, sentinel=_sentinel))
    guild: "Guild" = field(default=None)

    _guild_id: "Snowflake_Type" = field()
    _bot_id: Optional["Snowflake_Type"] = field(default=None)
    integration_id: Optional["Snowflake_Type"] = field(default=None)  # todo integration object?

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        data.update(data.pop("tags", {}))
        return data

    async def get_bot(self) -> Optional["Member"]:
        """
        Get the bot associated with this role if any.

        Returns:
            Member object if any
        """
        if self._bot_id is None:
            return None
        return await self._client.cache.get_member(self._guild_id, self._bot_id)

    @property
    def default(self) -> bool:
        """Is this the `@everyone` role"""
        return self.id == self._guild_id

    @property
    def bot_managed(self) -> bool:
        """Is this role owned/managed by a bot"""
        return self.bot_id is not None

    @property
    def integration(self) -> bool:
        """Is this role owned/managed by a integration"""
        return self.tags.integration_id is not None

    async def is_assignable(self) -> bool:
        """Can this role be assigned or removed by this bot?
        !!! note:
            This does not account for permissions, only the role hierarchy"""
        me = await self.guild.me

        if (self.default or await me.top_role.position > self.position) and not self.managed:
            return True
        return False

    async def delete(self, reason: str = None):
        """
        Delete this role

        Args:
            reason: An optional reason for this deletion
        """
        await self._client.http.delete_guild_role(self._guild_id, self.id, reason)

    async def edit(
        self,
        name: str = MISSING,
        permissions: str = MISSING,
        color: Union[int, Color] = MISSING,
        hoist: bool = MISSING,
        mentionable: bool = MISSING,
    ):
        """
        Edit this role, all arguments are optional.

        Args:
            name: name of the role
            permissions: New permissions to use
            color: The color of the role
            hoist: whether the role should be displayed separately in the sidebar
            mentionable: whether the role should be mentionable
        """

        if isinstance(color, Color):
            color = color.value

        payload = dict_filter_missing(
            dict(name=name, permissions=permissions, color=color, hoist=hoist, mentionable=mentionable)
        )

        r_data = await self._client.http.modify_guild_role(self._guild_id, self.id, payload)
        r_data["guild_id"] = self._guild_id
        return self.from_dict(r_data, self._client)



from base64 import b64encode
from ctypes import Union
from datetime import datetime, timezone
from io import IOBase
from pathlib import Path
from typing import Optional

from attr import fields, has

from dis_snek.const import MISSING


no_export_meta = dict(no_export=True)


def export_converter(converter) -> dict:
    return dict(export_converter=converter)


def to_dict(inst):
    if (converter := getattr(inst, "as_dict", None)) is not None:
        return converter()

    attrs = fields(inst.__class__)
    d = dict()

    for a in attrs:
        if a.metadata.get("no_export", False):
            continue

        raw_value = getattr(inst, a.name)
        if raw_value is MISSING:
            continue

        if (c := a.metadata.get("export_converter", None)) is not None:
            value = c(raw_value)
        else:
            value = _to_dict_any(raw_value)

        if isinstance(value, bool) or value:
            d[a.name] = value

    return d


def _to_dict_any(inst):
    if has(inst.__class__):
        return to_dict(inst)
    elif isinstance(inst, dict):
        return {key: _to_dict_any(value) for key, value in inst.items()}
    elif isinstance(inst, (list, tuple, set, frozenset)):
        return [_to_dict_any(item) for item in inst]
    elif isinstance(inst, datetime):
        if inst.tzinfo:
            return inst.isoformat()
        else:
            return inst.replace(tzinfo=timezone.utc).isoformat()
    else:
        return inst


def dict_filter_none(data: dict) -> dict:
    return {k: v for k, v in data.items() if v is not None}


def dict_filter_missing(data: dict) -> dict:
    return {k: v for k, v in data.items() if v is not MISSING}


def to_image_data(imagefile) -> Optional[str]:
    if isinstance(imagefile, IOBase):
        image_data = imagefile.read()
    elif isinstance(imagefile, (str, Path)):
        with open(imagefile, "rb") as image_buffer:
            image_data = image_buffer.read()
    else:
        return imagefile

    mimetype = _get_file_mimetype(image_data)
    encoded_image = b64encode(image_data).decode("ascii")

    return f"data:{mimetype};base64,{encoded_image}"


def _get_file_mimetype(filedata: bytes):
    if filedata.startswith((b"GIF87a", b"GIF89a")):
        return "image/gif"
    elif filedata.startswith(b"\x89PNG\x0D\x0A\x1A\x0A"):
        return "image/png"
    elif filedata.startswith(b"\xff\xd8\xff"):
        return "image/jpeg"
    elif filedata[0:4] == b"RIFF" and filedata[8:12] == b"WEBP":
        return "image/webp"
    else:
        return "application/octet-stream"



from typing import TYPE_CHECKING, Any, Set, Dict, List, Optional, Union

import attr
from attr.converters import optional as optional_c

from dis_snek.const import MISSING
from dis_snek.errors import HTTPException, TooManyChanges
from dis_snek.mixins.send import SendMixin
from dis_snek.models.color import Color
from dis_snek.models.discord import DiscordObject
from dis_snek.models.discord_objects.asset import Asset
from dis_snek.models.discord_objects.guild import Guild
from dis_snek.models.discord_objects.role import Role
from dis_snek.models.enums import Permissions, PremiumTypes, UserFlags
from dis_snek.models.snowflake import Snowflake_Type
from dis_snek.models.snowflake import to_snowflake
from dis_snek.models.timestamp import Timestamp
from dis_snek.utils.attr_utils import define, field, class_defaults
from dis_snek.utils.converters import list_converter
from dis_snek.utils.converters import timestamp_converter
from dis_snek.utils.input_utils import _bytes_to_base64_data

if TYPE_CHECKING:
    from aiohttp import FormData

    from dis_snek.client import Snake
    from dis_snek.models.discord_objects.channel import TYPE_GUILD_CHANNEL, DM


class _SendDMMixin(SendMixin):
    id: "Snowflake_Type"

    async def _send_http_request(self, message_payload: Union[dict, "FormData"]) -> dict:
        dm_id = await self._client.cache.get_dm_channel_id(self.id)
        return await self._client.http.create_message(message_payload, dm_id)


@define()
class BaseUser(DiscordObject, _SendDMMixin):
    """Base class for User, essentially partial user discord model"""

    username: str = field(repr=True, metadata={"docs": "The user's username, not unique across the platform"})
    discriminator: int = field(repr=True, metadata={"docs": "The user's 4-digit discord-tag"})
    avatar: "Asset" = field(metadata={"docs": "The user's default avatar"})

    def __str__(self):
        return f"{self.username}#{self.discriminator}"

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        data["avatar"] = Asset.from_path_hash(client, f"avatars/{data['id']}/{{}}", data["avatar"])
        return data

    @property
    def mention(self) -> str:
        """Returns a string that would mention the user"""
        return f"<@{self.id}>"

    @property
    def display_name(self) -> str:
        """The users display name, will return nickname if one is set, otherwise will return username"""
        return self.username  # for duck-typing compatibility with Member

    async def get_dm(self) -> "DM":
        """Get the DM channel associated with this user."""
        return await self._client.cache.get_channel(self.id)  # noqa

    @property
    def mutual_guilds(self) -> List["Guild"]:
        """Get a list of mutual guilds shared between this user and the client."""

        # should user_guilds be its own property?
        return [g for g in self._client.guilds if g.id in self.user_guilds]


@define()
class User(BaseUser):
    bot: bool = field(repr=True, default=False, metadata={"docs": "Is this user a bot?"})
    system: bool = field(
        default=False,
        metadata={"docs": "whether the user is an Official Discord System user (part of the urgent message system)"},
    )
    public_flags: "UserFlags" = field(
        repr=True, default=0, converter=UserFlags, metadata={"docs": "The flags associated with this user"}
    )
    premium_type: "PremiumTypes" = field(
        default=0, converter=PremiumTypes, metadata={"docs": "The type of nitro subscription on a user's account"}
    )

    banner: Optional["Asset"] = field(default=None, metadata={"docs": "The user's banner"})
    accent_color: Optional["Color"] = field(
        default=None,
        converter=optional_c(Color),
        metadata={"docs": "The user's banner color"},
    )

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        data = super()._process_dict(data, client)
        if "banner" in data:
            data["banner"] = Asset.from_path_hash(client, f"banners/{data['id']}/{{}}", data["banner"])

        return data


@define()
class SnakeBotUser(User):
    verified: bool = field(repr=True, metadata={"docs": ""})
    mfa_enabled: bool = field(default=False, metadata={"docs": ""})
    email: Optional[str] = field(default=None, metadata={"docs": ""})  # needs special permissions?
    locale: Optional[str] = field(default=None, metadata={"docs": ""})
    bio: Optional[str] = field(default=None, metadata={"docs": ""})
    flags: "UserFlags" = field(default=0, converter=UserFlags, metadata={"docs": ""})

    _guild_ids: Set[str] = field(factory=set, metadata={"docs": ""})

    def _add_guilds(self, guild_ids: Set["Snowflake_Type"]):
        self._guild_ids |= guild_ids

    @property
    def guilds(self) -> List["guilds"]:
        return [self._client.cache.guild_cache.get(g_id) for g_id in self._guild_ids]

    async def edit(self, username: Optional[str] = None, avatar: Optional[bytes] = MISSING):
        """
        Edit the client's user.

        You can either change the username, or avatar, or both at once.
        `avatar` may be set to `None` to remove your bot's avatar

        ??? Hint "Example Usage:"
            ```python
            f = open("path_to_file", "rb")
            await self.user.edit(avatar=f.read())
            ```
            or
            ```python
            await self.user.edit(username="hello world")
            ```
        Args:
            username: The username you want to use
            avatar: The avatar to use, must be `bytes` (see example)

        Returns:

        """
        payload = {}
        if username:
            payload["username"] = username
        if avatar:
            payload["avatar"] = _bytes_to_base64_data(avatar)  # noqa
        elif avatar is None:
            payload["avatar"] = None

        try:
            resp = await self._client.http.modify_client_user(payload)
        except HTTPException:
            raise TooManyChanges(
                "You have changed your profile too frequently, you need to wait a while before trying again."
            ) from None
        if resp:
            self._client.cache.place_user_data(resp)


@attr.s(**{k: v for k, v in class_defaults.items() if k != "on_setattr"})
class Member(DiscordObject, _SendDMMixin):
    bot: bool = field(repr=True, default=False, metadata={"docs": "Is this user a bot?"})
    nick: Optional[str] = field(repr=True, default=None, metadata={"docs": "The user's nickname in this guild'"})
    deaf: bool = field(default=False, metadata={"docs": "Has this user been deafened in voice channels?"})
    mute: bool = field(default=False, metadata={"docs": "Has this user been muted in voice channels?"})
    joined_at: "Timestamp" = field(converter=timestamp_converter, metadata={"docs": "When the user joined this guild"})
    premium_since: Optional["Timestamp"] = field(
        default=None,
        converter=optional_c(timestamp_converter),
        metadata={"docs": "When the user started boosting the guild"},
    )
    pending: Optional[bool] = field(
        default=None, metadata={"docs": "Whether the user has **not** passed guild's membership screening requirements"}
    )

    _guild_id: "Snowflake_Type" = field(repr=True, metadata={"docs": "The ID of the guild"})
    _role_ids: List["Snowflake_Type"] = field(
        factory=list, converter=list_converter(to_snowflake), metadata={"docs": "The roles IDs this user has"}
    )

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        if "user" in data:
            user_data = data.pop("user")
            client.cache.place_user_data(user_data)
            data["id"] = user_data["id"]
            data["bot"] = user_data.get("bot", False)
        elif "member" in data:
            member_data = data.pop("member")
            client.cache.place_user_data(data)
            member_data["id"] = data["id"]
            member_data["bot"] = data.get("bot", False)
            data = member_data

        data["role_ids"] = data.pop("roles", [])

        return data

    @property
    def user(self):
        """Returns this member's user object"""
        return self._client.cache.user_cache.get(self.id)

    def __getattr__(self, name):
        # this allows for transparent access to user attributes
        try:
            return getattr(self.user, name)
        except AttributeError:
            raise AttributeError(f"Neither `User` or `Member` have attribute {name}")

    def __setattr__(self, key, value):
        # this allows for transparent access to user attributes
        if attrib := getattr(self.__attrs_attrs__, key):
            value = attr.setters.convert(self, attrib, value)
            value = attr.setters.validate(self, attrib, value)
        super(Member, self).__setattr__(key, value)

    @property
    def nickname(self):
        """alias for nick"""
        return self.nick

    @nickname.setter
    def nickname(self, nickname):
        self.nick = nickname

    @property
    def guild(self) -> "Guild":
        return self._client.cache.guild_cache.get(self._guild_id)

    @property
    def roles(self) -> List["Role"]:
        return [r for r in self.guild.roles if r.id in self._role_ids]

    @property
    def top_role(self) -> "Role":
        return self._client.cache.role_cache.get(self._role_ids[-1])

    @property
    def display_name(self) -> str:
        """The users display name, will return nickname if one is set, otherwise will return username"""
        return self.nickname or self.username

    @property
    def premium(self) -> bool:
        """Is this member a server booster?"""
        return self.premium_since is not None

    async def guild_permissions(self) -> Permissions:
        """
        Returns the permissions this member has in the guild

        Returns:
            Permission data
        """
        guild = self.guild
        if guild.is_owner(self):
            return Permissions.ALL

        role_everyone = guild.default_role  # get @everyone role
        permissions = role_everyone.permissions

        for role in self.roles:
            permissions |= role.permissions

        if Permissions.ADMINISTRATOR in permissions:
            return Permissions.ALL

        return permissions

    async def has_permission(self, *permissions: Permissions) -> bool:
        """
        Checks if the member has all the given permission(s).

        ??? Hint "Example Usage:"
            Two different styles can be used to call this method.

            ```python
            await member.has_permission(Permissions.KICK_MEMBERS, Permissions.BAN_MEMBERS)
            ```
            ```python
            await member.has_permission(Permissions.KICK_MEMBERS | Permissions.BAN_MEMBERS)
            ```

            If `member` has both permissions, `True` gets returned.

        Args:
            permissions: The permission(s) to check whether the user has it.
        """

        # Get the user's permissions
        guild_permissions = await self.guild_permissions()

        # Check all permissions separately
        for permission in permissions:
            if permission not in guild_permissions:
                return False
        return True

    async def channel_permissions(self, channel: "TYPE_GUILD_CHANNEL") -> Permissions:
        """
        Returns the permissions this member has in a channel.

        Args:
            channel: The channel in question

        Returns:
            Permissions data
        """
        permissions = await self.guild_permissions()
        if Permissions.ADMINISTRATOR in permissions:
            return Permissions.ALL

        # Find (@everyone) role overwrite and apply it.
        overwrites = channel._permission_overwrites
        if overwrite_everyone := overwrites.get(channel._guild_id):
            permissions &= ~overwrite_everyone.deny
            permissions |= overwrite_everyone.allow

        # Apply role specific overwrites.
        allow = Permissions.NONE
        deny = Permissions.NONE
        for role_id in self.roles.ids:
            if overwrite_role := overwrites.get(role_id):
                allow |= overwrite_role.allow
                deny |= overwrite_role.deny

        permissions &= ~deny
        permissions |= allow

        # Apply member specific overwrite if it exist.
        if overwrite_member := overwrites.get(self.id):
            permissions &= ~overwrite_member.deny
            permissions |= overwrite_member.allow

        return permissions

    async def edit_nickname(self, new_nickname: str):
        """
        Change the user's nickname.

        Args:
            new_nickname: The new nickname to apply.
        """
        return await self._client.http.modify_guild_member(self._guild_id, self.id, nickname=new_nickname)

    async def add_role(self, role: Union[Snowflake_Type, Role], reason: str = MISSING):
        """
        Add a role to this member.
        Args:
            role: The role to add
            reason: The reason for adding this role
        """
        role = to_snowflake(role)
        return await self._client.http.add_guild_member_role(self._guild_id, self.id, role, reason=reason)

    async def remove_role(self, role: Union[Snowflake_Type, Role], reason: str = MISSING):
        """
        Remove a role from this user.
        Args:
            role: The role to remove
            reason: The reason for this removal
        """
        if isinstance(role, Role):
            role = role.id
        return await self._client.http.remove_guild_member_role(self._guild_id, self.id, role, reason=reason)

    async def has_role(self, *roles: Union[Snowflake_Type, Role]) -> bool:
        """
        Checks if the user has the given role(s)
        Args:
            roles: The role(s) to check whether the user has it.
        """
        for role in roles:
            role_id = to_snowflake(role)
            if role_id not in self._role_ids:
                return False
        return True

    async def kick(self, reason: str = MISSING):
        """
        Remove a member from the guild.
        Args:
            reason: The reason for this removal
        """
        return await self._client.http.remove_guild_member(self._guild_id, self.id)

    async def ban(self, delete_message_days=0, reason: str = MISSING):
        """
        Ban a member from the guild.
        Args:
            delete_message_days: The number of days of messages to delete
            reason: The reason for this ban
        """
        return await self._client.http.create_guild_ban(self._guild_id, self.id, delete_message_days, reason=reason)



import asyncio
import inspect
import logging
from typing import List, TYPE_CHECKING, Callable, Coroutine

from dis_snek.const import logger_name
from dis_snek.models.application_commands import InteractionCommand
from dis_snek.models.command import MessageCommand
from dis_snek.models.listener import Listener
from dis_snek.utils.misc_utils import wrap_partial

if TYPE_CHECKING:
    from dis_snek.client import Snake

log = logging.getLogger(logger_name)


class Scale:
    """
    A class that allows you to separate your commands and listeners into separate files.
    Skins require an entrypoint in the same file called `setup`, this function allows
    client to load the Scale.

    ??? Hint "Example Usage:"
        ```python
        class ExampleScale(Scale):
            def __init__(self, bot):
                print("Scale Created")

            @message_command
            async def some_command(self, context):
                await ctx.send(f"I was sent from a scale called {self.name}")
        ```

    parameters:
        bot Snake: A reference to the client

    Attributes:
        bot Snake: A reference to the client
        name str: The name of this Scale (`read-only`)
        description str: A description of this Scale
        scale_checks str: A list of checks to be ran on any command in this scale
        scale_prerun List: A list of coroutines to be run before any command in this scale
        scale_postrun List: A list of coroutines to be run after any command in this scale
    """

    bot: "Snake"
    _commands: List
    __name: str
    extension_name: str
    description: str
    scale_checks: List
    scale_prerun: List
    scale_postrun: List
    listeners: List

    def __new__(cls, bot: "Snake", *args, **kwargs):
        cls.bot = bot
        cls.__name = cls.__name__
        cls.scale_checks = []
        cls.scale_prerun = []
        cls.scale_postrun = []
        cls.listeners = []

        cls.description = kwargs.get("Description", None)
        if not cls.description:
            cls.description = inspect.cleandoc(cls.__doc__) if cls.__doc__ else None

        # load commands from class
        cls._commands = []
        cls._listeners = []

        new_cls = super().__new__(cls)

        for name, val in cls.__dict__.items():
            if isinstance(val, (InteractionCommand, MessageCommand)):
                val.scale = new_cls
                val = wrap_partial(val, new_cls)

                new_cls._commands.append(val)

                if isinstance(val, InteractionCommand):
                    bot.add_interaction(val)
                else:
                    bot.add_message_command(val)
            elif isinstance(val, Listener):
                val = wrap_partial(val, new_cls)
                bot.add_listener(val)
                new_cls.listeners.append(val)

        log.debug(
            f"{len(new_cls._commands)} commands and {len(new_cls.listeners)} listeners"
            f" have been loaded from `{new_cls.name}`"
        )

        new_cls.extension_name = inspect.getmodule(new_cls).__name__
        new_cls.bot.scales[new_cls.name] = new_cls
        return new_cls

    @property
    def __name__(self):
        return self.name

    @property
    def commands(self):
        """Get the commands from this Scale"""
        return self._commands

    @property
    def name(self):
        return self.__name

    def shed(self):
        """
        Called when this Scale is being removed.
        """
        for func in self._commands:
            if isinstance(func, InteractionCommand):
                if self.bot.interactions.get(func.scope) and self.bot.interactions[func.scope].get(func.name):
                    self.bot.interactions[func.scope].pop(func.name)
            if isinstance(func, MessageCommand):
                if self.bot.commands[func.name]:
                    self.bot.commands.pop(func.name)
        for func in self.listeners:
            self.bot.listeners[func.event].remove(func)

        self.bot.scales.pop(self.name, None)
        log.debug(f"{self.name} has been shed")

    def add_scale_check(self, coroutine: Callable[..., Coroutine]) -> None:
        """
        Add a coroutine as a check for all commands in this scale to run. This coroutine must take **only** the parameter `context`.

        ??? Hint "Example Usage:"
            ```python
            def __init__(self, bot):
                self.bot = bot
                self.add_scale_check(self.example)

            @staticmethod
            async def example(context: Context):
                if context.author.id == 123456789:
                    return True
                return False
            ```
        Args:
            coroutine: The coroutine to use as a check
        """
        if not asyncio.iscoroutinefunction(coroutine):
            raise TypeError("Check must be a coroutine")

        if not self.scale_checks:
            self.scale_checks = []

        self.scale_checks.append(coroutine)

    def add_scale_prerun(self, coroutine: Callable[..., Coroutine]):
        """
        Add a coroutine to be run **before** all commands in this Scale.

        Note:
            Pre-runs will **only** be run if the commands checks pass

        ??? Hint "Example Usage:"
            ```python
            def __init__(self, bot):
                self.bot = bot
                self.add_scale_prerun(self.example)

            async def example(self, context: Context):
                await ctx.send("I ran first")
            ```

        Args:
            coroutine: The coroutine to run
        """
        if not asyncio.iscoroutinefunction(coroutine):
            raise TypeError("Callback must be a coroutine")

        if not self.scale_prerun:
            self.scale_prerun = []
        self.scale_prerun.append(coroutine)

    def add_scale_postrun(self, coroutine: Callable[..., Coroutine]):
        """
        Add a coroutine to be run **after** all commands in this Scale.

        ??? Hint "Example Usage:"
            ```python
            def __init__(self, bot):
                self.bot = bot
                self.add_scale_postrun(self.example)

            async def example(self, context: Context):
                await ctx.send("I ran first")
            ```

        Args:
            coroutine: The coroutine to run
        """
        if not asyncio.iscoroutinefunction(coroutine):
            raise TypeError("Callback must be a coroutine")

        if not self.scale_postrun:
            self.scale_postrun = []
        self.scale_postrun.append(coroutine)



from typing import TYPE_CHECKING, List, Optional, Dict, Any

import attr
from attr.converters import optional

from dis_snek.models.discord import DiscordObject
from dis_snek.models.enums import ApplicationFlags
from dis_snek.models.snowflake import Snowflake_Type, to_snowflake

if TYPE_CHECKING:
    from dis_snek.client import Snake
    from dis_snek.models.discord_objects.team import Team


@attr.s(slots=True, kw_only=True)
class Application(DiscordObject):
    """
    Represents a discord application.
    Attributes:
        name str: The name of the application.
        icon str: The icon of the application
        description str: The description of the application
        rpc_origins List[str]: an array of rpc origin urls, if rpc is enabled
        bot_public bool: when false only app owner can join the app's bot to guilds
        bot_require_code_grant bool: when true the app's bot will only join upon completion of the full oauth2 code grant flow
        terms_of_service_url str: the url of the app's terms of service
        privacy_policy_url str: the url of the app's privacy policy
        owner [User][dis_snek.models.discord_objects.user.BaseUser] : partial user object containing info on the owner of the application
        summary str: if this application is a game sold on Discord, this field will be the summary field for the store page of its primary sku
        verify_key str: the hex encoded key for verification in interactions and the GameSDK's GetTicket
        team Team: if the application belongs to a team, this will be a list of the members of that team
        guild_id Snowflake_Type: if this application is a game sold on Discord, this field will be the guild to which it has been linked
        primary_sku_id Snowflake_Type: if this application is a game sold on Discord, this field will be the id of the "Game SKU" that is created, if exists
        slug str: if this application is a game sold on Discord, this field will be the URL slug that links to the store page
        cover_image str: the application's default rich presence invite cover image hash
        flags int: the application's public flags
    """

    name: str = attr.ib()
    icon: Optional[str] = attr.ib(default=None)
    description: Optional[str] = attr.ib()
    rpc_origins: Optional[List[str]] = attr.ib(default=None)
    bot_public: bool = attr.ib(default=True)
    bot_require_code_grant: bool = attr.ib(default=False)
    terms_of_service_url: Optional[str] = attr.ib(default=None)
    privacy_policy_url: Optional[str] = attr.ib(default=None)
    owner_id: Snowflake_Type = attr.ib(converter=to_snowflake)
    summary: str = attr.ib()
    verify_key: str = attr.ib()
    team: Optional["Team"] = attr.ib(default=None)
    guild_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    primary_sku_id: Optional["Snowflake_Type"] = attr.ib(default=None)
    slug: Optional[str] = attr.ib(default=None)
    cover_image: Optional[str] = attr.ib(default=None)
    flags: Optional["ApplicationFlags"] = attr.ib(default=None, converter=optional(ApplicationFlags))

    @classmethod
    def _process_dict(cls, data: Dict[str, Any], client: "Snake") -> Dict[str, Any]:
        owner = client.cache.place_user_data(data.pop("owner"))
        data["owner_id"] = owner.id
        return data

    @property
    def owner(self):
        return self._client.cache.user_cache.get(self.owner_id)
